"""Tidy3d user account."""

from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic.v1 import Extra, Field

from .http_util import http
from .types import Tidy3DResource


class Account(Tidy3DResource, extra=Extra.allow):
    """Tidy3D User Account."""

    allowance_cycle_type: Optional[str] = Field(
        None,
        title="AllowanceCycleType",
        description="Daily or Monthly",
        alias="allowanceCycleType",
    )
    credit: Optional[float] = Field(
        0, title="credit", description="Current FlexCredit balance", alias="credit"
    )
    credit_expiration: Optional[datetime] = Field(
        None,
        title="creditExpiration",
        description="Expiration date",
        alias="creditExpiration",
    )
    allowance_current_cycle_amount: Optional[float] = Field(
        0,
        title="allowanceCurrentCycleAmount",
        description="Daily/Monthly free simulation balance",
        alias="allowanceCurrentCycleAmount",
    )
    allowance_current_cycle_end_date: Optional[datetime] = Field(
        None,
        title="allowanceCurrentCycleEndDate",
        description="Daily/Monthly free simulation balance expiration date",
        alias="allowanceCurrentCycleEndDate",
    )
    daily_free_simulation_counts: Optional[int] = Field(
        0,
        title="dailyFreeSimulationCounts",
        description="Daily free simulation counts",
        alias="dailyFreeSimulationCounts",
    )

    @classmethod
    def get(cls):
        """Get user account information.

        Parameters
        ----------

        Returns
        -------
        account : Account
        """
        resp = http.get("tidy3d/py/account")
        if resp:
            account = Account(**resp)
            return account
        else:
            return None



from typing import Union

from .multi_physics import MultiPhysicsMedium
from .solver_types import (
    ChargeMediumType,
    ChargeMediumType3D,
    ElectricalMediumType,
    ElectricalMediumType3D,
    HeatMediumType,
    OpticalMediumType,
    OpticalMediumType3D,
)

StructureMediumType = Union[
    MultiPhysicsMedium,
    OpticalMediumType,
    ElectricalMediumType,
    HeatMediumType,
    ChargeMediumType,
]

MultiPhysicsMediumType3D = Union[
    MultiPhysicsMedium, OpticalMediumType3D, ElectricalMediumType3D, ChargeMediumType3D
]



"""Classes for creating data based on analytic beams like plane wave, Gaussian beam, and
astigmatic Gaussian beam."""

from abc import abstractmethod
from typing import Optional, Tuple, Union

import autograd.numpy as np
import pydantic.v1 as pd

from ..constants import C_0, ETA_0, HERTZ, MICROMETER, RADIAN
from .base import cached_property
from .data.data_array import ScalarFieldDataArray
from .data.monitor_data import FieldData
from .geometry.base import Box
from .grid.grid import Coords, Grid
from .medium import Medium, MediumType
from .monitor import FieldMonitor
from .source.field import FixedAngleSpec, FixedInPlaneKSpec
from .types import TYPE_TAG_STR, Direction, FreqArray, Literal, Numpy
from .validators import assert_plane

DEFAULT_RESOLUTION = 200


class BeamProfile(Box):
    """Base class for handling analytic beams."""

    resolution: float = pd.Field(
        DEFAULT_RESOLUTION,
        title="Sampling resolution",
        description="Sampling resolution in the tangential directions of the beam (defines a "
        "number of equally spaced points).",
        units=MICROMETER,
    )

    freqs: FreqArray = pd.Field(
        ...,
        title="Frequencies",
        description="List of frequencies at which the beam is sampled.",
        units=HERTZ,
    )

    background_medium: MediumType = pd.Field(
        Medium(),
        title="Background Medium",
        description="Background medium in which the beam is embedded.",
    )

    angle_theta: float = pd.Field(
        0.0,
        title="Polar Angle",
        description="Polar angle of the propagation axis from the normal axis.",
        units=RADIAN,
    )

    angle_phi: float = pd.Field(
        0.0,
        title="Azimuth Angle",
        description="Azimuth angle of the propagation axis in the plane orthogonal to the "
        "normal axis.",
        units=RADIAN,
    )

    pol_angle: float = pd.Field(
        0.0,
        title="Polarization Angle",
        description="Specifies the angle between the electric field polarization of the "
        "beam and the plane defined by the normal axis and the propagation axis (rad). "
        "``pol_angle=0`` (default) specifies P polarization, "
        "while ``pol_angle=np.pi/2`` specifies S polarization. "
        "At normal incidence when S and P are undefined, ``pol_angle=0`` defines: "
        "- ``Ey`` polarization for propagation along ``x``."
        "- ``Ex`` polarization for propagation along ``y``."
        "- ``Ex`` polarization for propagation along ``z``.",
        units=RADIAN,
    )

    direction: Direction = pd.Field(
        "+",
        title="Direction",
        description="Specifies propagation in the positive or negative direction of the normal "
        "axis.",
    )

    _plane_validator = assert_plane()

    @property
    def grid(self) -> Grid:
        """Return a Grid object on which the beam will be sampled."""
        dim_n, dim_tan = self.pop_axis("xyz", self.size.index(0.0))
        bounds_n, bounds_tan = self.pop_axis(np.array(self.bounds).T, self.size.index(0.0))
        boundaries = {
            dim_n: bounds_n,
            dim_tan[0]: np.linspace(bounds_tan[0][0], bounds_tan[0][1], int(self.resolution)),
            dim_tan[1]: np.linspace(bounds_tan[1][0], bounds_tan[1][1], int(self.resolution)),
        }
        return Grid(boundaries=boundaries)

    @property
    def monitor(self) -> FieldMonitor:
        """``FieldMonitor`` with the same center, size, and frequencies as the beam, to be used
        in the ``FieldData`` created by the ``field_data`` method."""
        return FieldMonitor(
            size=self.size,
            center=self.center,
            freqs=self.freqs,
            name="<<BEAM_DATA>>",
        )

    @cached_property
    def field_data(self) -> FieldData:
        """Compute a FieldData for the spatial E and H field amplitudes of an analytically defined
        beam."""

        background_n = np.zeros(len(self.freqs), dtype=complex)
        for freq_id, freq in enumerate(self.freqs):
            eps = self.background_medium.eps_model(freq)
            nk_medium = self.background_medium.eps_complex_to_nk(eps)
            background_n[freq_id] = np.squeeze(nk_medium[0]) + 1j * np.squeeze(nk_medium[1])

        data_dict = self._field_data_on_grid(grid=self.grid, background_n=background_n)
        data_raw = FieldData(monitor=self.monitor, grid_expanded=self.grid, **data_dict)

        # Normalize by the flux
        fields_norm = {}
        flux = np.abs(data_raw.flux)
        for field_name, field_data in data_raw.field_components.items():
            fields_norm[field_name] = (field_data / np.sqrt(flux)).astype(field_data.dtype)

        return data_raw.updated_copy(**fields_norm)

    def _field_data_on_grid(self, grid: Grid, background_n: Numpy, colocate=True) -> dict:
        """Compute the field data for each field component on a grid for the beam.
        A dictionary of the scalar field data arrays is returned, not yet packaged as ``FieldData``.
        """

        field_components = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]
        if colocate:
            # Just sample at the grid boundaries for each field component
            # We skip the last boundary because that's how things also work in the solver (no
            # field defined at the last boundary)
            coords_dict = {key: val[:-1] for key, val in grid.boundaries.to_dict.items()}
            grid_dict = {component: Coords(**coords_dict) for component in field_components}
        else:
            # Yee grid for each field component
            grid_dict = grid.yee.grid_dict

        # Compute each field component over the beam grid
        scalar_fields = {}
        for comp, field in enumerate(field_components):
            x, y, z = grid_dict[field].to_list
            # Center component grid at beam center
            x_c, y_c, z_c = (coords - cent for coords, cent in zip((x, y, z), self.center))
            # Stack E and H fields
            field_vals = self.analytic_beam(x_c, y_c, z_c, background_n, field=field[0])
            # Get the current field component
            field_vals = field_vals[comp % 3]
            # Make the ScalarFieldDataArray for the current component
            coords = dict(x=x, y=y, z=z, f=np.array(self.freqs))
            field_data = ScalarFieldDataArray(field_vals, coords=coords)
            scalar_fields[field] = field_data

        return scalar_fields

    @abstractmethod
    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``."""
        pass

    def analytic_beam_z_normal(
        self, points: Numpy, background_n: float, field: Literal["E", "H"]
    ) -> Numpy:
        """Analytic beam with all the beam parameters but assuming ``z`` as the normal axis."""

        # Add a frequency dimension to points
        points = np.repeat(points[:, :, np.newaxis], len(self.freqs), axis=2)

        # Rotate points to axes where propagation is along z
        points_prop_z = self._rotate_points_z(points, background_n)

        # Reflection at the z = 0 plane for negative direction
        if self.direction == "-":
            points_prop_z[2, ...] *= -1

        # Get fields polarized along x and propagating along z
        scalar_field = self.scalar_field(points_prop_z, background_n)

        # Set the correct field component based on the scalar field values
        field_vals = np.zeros(points.shape, dtype=np.complex128)
        if field == "E":
            field_vals[0, ...] = scalar_field
        else:
            field_vals[1, ...] = scalar_field / ETA_0 * background_n

        # Rotate polarization
        field_vals = self.rotate_points(field_vals, [0, 0, 1], self.pol_angle)

        # Reflection of the field components at the z = 0 plane for negative direction
        if self.direction == "-":
            if field == "E":
                field_vals[2, :] *= -1
            else:
                field_vals[:2, :] *= -1

        # Rotate the fields back to the original propagation axes
        field_vals = self._inverse_rotate_field_vals_z(field_vals, background_n)

        return field_vals

    def analytic_beam(
        self,
        x: Numpy,
        y: Numpy,
        z: Numpy,
        background_n: float,
        field: Literal["E", "H"],
    ) -> Numpy:
        """Sample the analytic beam fields on a cartesian grid of points in x, y, z."""

        # Make a meshgrid
        (x_mesh, y_mesh, z_mesh) = np.meshgrid(x, y, z, indexing="ij")
        (Nx, Ny, Nz) = x_mesh.shape

        # Move to coordinates where the normal axis is z
        (z_new, (x_new, y_new)) = self.pop_axis((x_mesh, y_mesh, z_mesh), axis=self.size.index(0.0))
        points = np.stack([coords.ravel() for coords in (x_new, y_new, z_new)], axis=0)

        # Sample the plane wave fields
        f_x, f_y, f_z = self.analytic_beam_z_normal(points, background_n, field)

        # Move back to original coordinates
        field_vals = np.stack(self.unpop_axis(f_z, (f_x, f_y), axis=self.size.index(0.0)), axis=0)

        # H field gets a -1 factor if a reflection was involved
        if self.size.index(0.0) == 1 and field == "H":
            field_vals *= -1

        # Reshape to (3, Nx, Ny, Nz, num_freqs)
        return np.reshape(field_vals, (3, Nx, Ny, Nz, len(self.freqs)))

    def _rotate_points_z(self, points: Numpy, background_n: Numpy) -> Numpy:
        """Rotate points to new coordinates where z is the propagation axis."""
        points_prop_z = self.rotate_points(points, [0, 0, 1], -self.angle_phi)
        points_prop_z = self.rotate_points(points_prop_z, [0, 1, 0], -self.angle_theta)
        return points_prop_z

    def _inverse_rotate_field_vals_z(self, field_vals: Numpy, background_n: Numpy) -> Numpy:
        """Rotate field values from coordinates where z is the propagation axis to angled
        coordinates."""
        field_vals = self.rotate_points(field_vals, [0, 1, 0], self.angle_theta)
        field_vals = self.rotate_points(field_vals, [0, 0, 1], self.angle_phi)
        return field_vals


class PlaneWaveBeamProfile(BeamProfile):
    """Component for constructing plane wave beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.PlaneWave`.
    """

    angular_spec: Union[FixedInPlaneKSpec, FixedAngleSpec] = pd.Field(
        FixedAngleSpec(),
        title="Angular Dependence Specification",
        description="Specification of plane wave propagation direction dependence on wavelength.",
        discriminator=TYPE_TAG_STR,
    )

    as_fixed_angle_source: bool = pd.Field(
        False,
        title="Fixed Angle Flag",
        description="Fixed angle flag. Only used internally when computing source beams for "
        "injection in an FDTD simulation with fixed angle boudnaries. Use ``angular_spec`` to "
        "switch between waves with fixed angle and fixed in-plane k.",
    )

    angle_theta_frequency: Optional[float] = pd.Field(
        None,
        title="Frequency at Which Angle Theta is Defined",
        description="Frequency for which ``angle_theta`` is set. This only has an effect for "
        "fixed in-plane wave-vector beams. If not supplied, the average of the beam ``freqs`` is "
        "used.",
    )

    @property
    def _angle_theta_frequency(self):
        if not self.angle_theta_frequency:
            return np.mean(self.freqs)
        return self.angle_theta_frequency

    def in_plane_k(self, background_n: float):
        """In-plane wave vector. Only the real part is taken so the beam has no in-plane decay."""
        k0 = 2 * np.pi * self._angle_theta_frequency / C_0 * background_n
        k_in_plane = k0.real * np.sin(self.angle_theta)
        return [k_in_plane * np.cos(self.angle_phi), k_in_plane * np.sin(self.angle_phi)]

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field for plane wave.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, N_points, N_freqs)``.
        For the special case of fixed in-plane k, the propagation axis is different at every
        frequency, and the points a frquency-dependent rotation has been applied to the
        ``points`` in ``self._rotate_points_z``.
        """

        kz = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        if self.as_fixed_angle_source:
            kz *= np.cos(self.angle_theta)
        return np.exp(1j * points[2] * kz)

    def _angle_theta_actual(self, background_n: Numpy) -> Numpy:
        """Compute the frequency-dependent actual propagation angle theta."""
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        kx, ky = self.in_plane_k(background_n)
        k_perp = np.sqrt(kx**2 + ky**2)
        return np.real(np.arcsin(k_perp / k0))

    def _rotate_points_z(self, points: Numpy, background_n: Numpy) -> Numpy:
        """Rotate points to new coordinates where z is the propagation axis."""
        if self.as_fixed_angle_source:
            # For fixed-angle, we do not rotate the points
            return points
        elif isinstance(self.angular_spec, FixedInPlaneKSpec):
            # For fixed in-plane k, the rotation is angle-dependent
            points = self.rotate_points(points, [0, 0, 1], -self.angle_phi)
            angle_theta_actual = self._angle_theta_actual(background_n=background_n)
            for ind, theta_actual in enumerate(angle_theta_actual):
                points[:, :, ind] = self.rotate_points(points[:, :, ind], [0, 1, 0], -theta_actual)
            return points
        return super()._rotate_points_z(points, background_n)

    def _inverse_rotate_field_vals_z(self, field_vals: Numpy, background_n: Numpy) -> Numpy:
        """Rotate field values from coordinates where z is the propagation axis to angled
        coordinates. Special handling is needed if fixed in-plane k wave."""
        if isinstance(self.angular_spec, FixedInPlaneKSpec):
            # For fixed in-plane k, the rotation is angle-dependent
            angle_theta_actual = self._angle_theta_actual(background_n=background_n)
            for ind, theta_actual in enumerate(angle_theta_actual):
                field_vals[:, :, ind] = self.rotate_points(
                    field_vals[:, :, ind], [0, 1, 0], theta_actual
                )
            field_vals = self.rotate_points(field_vals, [0, 0, 1], self.angle_phi)
            return field_vals
        return super()._inverse_rotate_field_vals_z(field_vals, background_n)


class GaussianBeamProfile(BeamProfile):
    """Component for constructing Gaussian beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.GaussianBeam`.
    """

    waist_radius: pd.PositiveFloat = pd.Field(
        1.0,
        title="Waist Radius",
        description="Radius of the beam at the waist.",
        units=MICROMETER,
    )

    waist_distance: float = pd.Field(
        0.0,
        title="Waist Distance",
        description="Distance from the beam waist along the propagation direction. "
        "A positive value means the waist is positioned behind the beam, considering the propagation direction. "
        "For example, for a beam propagating in the ``+`` direction, a positive value of ``beam_distance`` "
        "means the beam waist is positioned in the ``-`` direction (behind the beam). "
        "A negative value means the beam waist is in the ``+`` direction (in front of the beam). "
        "For an angled beam, the distance is defined along the rotated propagation direction.",
        units=MICROMETER,
    )

    def beam_params(self, z: Numpy, k0: Numpy) -> Tuple[Numpy, Numpy, Numpy]:
        """Compute the parameters needed to evaluate a Gaussian beam at z.

        Parameters
        ----------
        z : Numpy
            Axial distance from the beam center.
        k0 : Numpy
            Wave vector magnitude.
        """

        w_0, z_0 = self.waist_radius, self.waist_distance
        z_r = w_0**2 * k0 / 2  # shape k0
        w_z = w_0 * np.sqrt(1 + ((z + z_0) / z_r) ** 2)  # shape (Np, Nk0)
        # inv_r_z shape (Np, Nk0)
        inv_r_z = (z + z_0) / ((z + z_0) ** 2 + z_r**2)
        # we choose gauge such that psi_g = 0 at z = 0 (beam plane)
        # this is needed for a proper interpolation between different frequencies
        # psi_g shape (Np, Nk0)
        psi_g = np.arctan((z + z_0) / z_r) - np.arctan(z_0 / z_r)
        return w_z, inv_r_z, psi_g

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """Scalar field for Gaussian beam.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``.
        """
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        x, y, z = points
        w_0 = self.waist_radius
        w_z, inv_r_z, psi_g = self.beam_params(z, k0)
        r_2 = x**2 + y**2
        scalar_gaussian = w_0 / w_z
        scalar_gaussian *= np.exp(-r_2 / w_z**2)
        scalar_gaussian *= np.exp(1j * (z * k0 + r_2 * k0 / 2 * inv_r_z - psi_g))

        return scalar_gaussian


class AstigmaticGaussianBeamProfile(BeamProfile):
    """Component for constructing astigmatic Gaussian beam data. The normal direction is implicitly
    defined by the ``size`` parameter.

    See also :class:`.AstigmaticGaussianBeam`.
    """

    waist_sizes: Tuple[pd.PositiveFloat, pd.PositiveFloat] = pd.Field(
        (1.0, 1.0),
        title="Waist sizes",
        description="Size of the beam at the waist in the local x and y directions.",
        units=MICROMETER,
    )

    waist_distances: Tuple[float, float] = pd.Field(
        (0.0, 0.0),
        title="Waist distances",
        description="Distance to the beam waist along the propagation direction "
        "for the waist sizes in the local x and y directions. "
        "When ``direction`` is ``+`` and ``waist_distances`` are positive, the waist "
        "is on the ``-`` side (behind) the beam plane. When ``direction`` is ``+`` and "
        "``waist_distances`` are negative, the waist is on the ``+`` side (in front) of "
        "the beam plane.",
        units=MICROMETER,
    )

    def beam_params(self, z: Numpy, k0: Numpy) -> Tuple[Numpy, Numpy, Numpy, Numpy]:
        """Compute the parameters needed to evaluate an astigmatic Gaussian beam at z.

        Parameters
        ----------
        z : Numpy
            Axial distance from the beam center.
        k0 : Numpy
            Wave vector magnitude.
        """

        w_xy, z_xy = self.waist_sizes, self.waist_distances  # shape (2, )
        z_r = [w**2 * k0 / 2 for w in w_xy]  # shape (2, Nk0)
        w_z, w_0, inv_r_z, psi_g = [], [], [], []  # final shape (2, Np, Nk0) after loop below
        for w, z_i, z_ri in zip(w_xy, z_xy, z_r):
            w_z.append(w * np.sqrt(1 + ((z + z_i) / z_ri) ** 2))
            w_0.append(w * np.sqrt(1 + (z_i / z_ri) ** 2))
            inv_r_z.append((z + z_i) / ((z + z_i) ** 2 + z_ri**2))
            # we choose gauge such that psi_g = 0 at z = 0 (beam plane)
            # this is needed for a proper interpolation between different frequencies
            # psi_g shape (Np, Nk0)
            psi_g_1 = np.arctan((z + z_i) / z_ri) / 2
            psi_g_2 = np.arctan(z_i / z_ri) / 2
            psi_g.append(psi_g_1 - psi_g_2)

        return w_0, w_z, inv_r_z, psi_g

    def scalar_field(self, points: Numpy, background_n: float) -> Numpy:
        """
        Scalar field for astigmatic Gaussian beam.
        Scalar field corresponding to the analytic beam in coordinate system such that the
        propagation direction is z and the ``E``-field is entirely ``x``-polarized. The field is
        computed on an unstructured array ``points`` of shape ``(3, ...)``.
        """
        k0 = 2 * np.pi * np.array(self.freqs) / C_0 * background_n
        x, y, z = points
        w_0, w_z, inv_r_z, psi_g = self.beam_params(z, k0)
        x_2 = x**2
        y_2 = y**2

        q_term1_inv = 1j * x_2 * k0 / 2 * inv_r_z[0] - x_2 / w_z[0] ** 2
        q_term2_inv = 1j * y_2 * k0 / 2 * inv_r_z[1] - y_2 / w_z[1] ** 2
        angle_term = q_term1_inv + q_term2_inv + 1j * (z * k0 - psi_g[0] - psi_g[1])
        scalar_gaussian = np.exp(angle_term)

        ampl = np.sqrt(w_0[0] * w_0[1] / w_z[0] / w_z[1])
        return scalar_gaussian * ampl



"""Parametric material models."""

import warnings
from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import pydantic.v1 as pd

from ..components.base import Tidy3dBaseModel
from ..components.medium import Drude, Medium2D, PoleResidue
from ..constants import ELECTRON_VOLT, EPSILON_0, HBAR, K_B, KELVIN, Q_e
from ..log import log

try:
    from scipy import integrate

    INTEGRATE_AVAILABLE = True
except ImportError:
    INTEGRATE_AVAILABLE = False

# default values of the physical parameters for graphene
# scattering rate in eV
GRAPHENE_DEF_GAMMA = 0.00041
# chemical potential in eV
GRAPHENE_DEF_MU_C = 0
# temperature in K
GRAPHENE_DEF_TEMP = 300

# constants controlling the numerical integration of the interband term in graphene
# frequency limits of integration
GRAPHENE_INT_MIN = 1e10
GRAPHENE_INT_MAX = 1e16
# integration absolute tolerance
GRAPHENE_INT_TOL = 1e-20

# constants controlling the Pade approximation of the interband term in graphene
# frequency range for fitting
GRAPHENE_FIT_FREQ_MIN = 1e12
GRAPHENE_FIT_FREQ_MAX = 1e15
GRAPHENE_FIT_NUM_FREQS = 100
GRAPHENE_FIT_ATOL = 2e-5
# parameters controlling node placement
GRAPHENE_FIT_LARGE_MULTIPLIER = 10
GRAPHENE_FIT_SMALL_MULTIPLIER = 0.25
# number of optimization iterations for fitting
GRAPHENE_FIT_NUM_ITERS = 100


class ParametricVariantItem2D(ABC, Tidy3dBaseModel):
    """A variant of a 2D material depending on parameters, that must be initialized in order to
    generate the material model."""

    @property
    @abstractmethod
    def medium(self) -> Medium2D:
        """Calculate the material model at the current parameter values."""


class Graphene(ParametricVariantItem2D):
    """Parametric surface conductivity model for graphene.

    Note
    ----
    The model contains intraband and interband terms, as described in::

        George W. Hanson, "Dyadic Green's Functions for an Anisotropic,
        Non-Local Model of Biased Graphene," IEEE Trans. Antennas Propag.
        56, 3, 747-757 (2008).

    Example
    -------
    >>> graphene_medium = Graphene(mu_c = 0.2).medium

    """

    mu_c: float = pd.Field(
        GRAPHENE_DEF_MU_C,
        title="Chemical potential in eV",
        description="Chemical potential in eV.",
        units=ELECTRON_VOLT,
    )
    temp: float = pd.Field(
        GRAPHENE_DEF_TEMP, title="Temperature in K", description="Temperature in K.", units=KELVIN
    )
    gamma: float = pd.Field(
        GRAPHENE_DEF_GAMMA,
        title="Scattering rate in eV",
        description="Scattering rate in eV. Must be small compared to the optical frequency.",
        units=ELECTRON_VOLT,
    )
    scaling: float = pd.Field(
        1,
        title="Scaling factor",
        description="Scaling factor used to model multiple layers of graphene.",
    )

    include_interband: bool = pd.Field(
        True,
        title="Include interband terms",
        description="Include interband terms, relevant at high frequency (IR). "
        "Otherwise, the intraband terms only give a simpler Drude-type model relevant "
        "only at low frequency (THz).",
    )
    interband_fit_freq_nodes: List[Tuple[float, float]] = pd.Field(
        None,
        title="Interband fitting frequency nodes",
        description="Frequency nodes for fitting interband term. "
        "Each pair of nodes in the list corresponds to a single Pade approximant of order "
        "(1, 2), which is optimized to minimize the error at these two frequencies. "
        "The default behavior is to fit a first approximant at one very low frequency and "
        "one very high frequency, and to fit a second approximant in the vicinity of the "
        "interband feature. This default behavior works for a wide range "
        "of frequencies; consider changing the nodes to obtain a better fit for a "
        "narrow-band simulation.",
    )
    interband_fit_num_iters: pd.NonNegativeInt = pd.Field(
        GRAPHENE_FIT_NUM_ITERS,
        title="Interband fitting number of iterations",
        description="Number of iterations for optimizing each Pade approximant when "
        "fitting the interband term. Making this larger might give a better fit "
        "at the cost of decreased stability in the fitting algorithm.",
    )

    @property
    def medium(self) -> Medium2D:
        """Surface conductivity model for graphene."""
        intraband = self.intraband_drude
        if self.include_interband:
            interband = self.interband_pole_residue
            intraband_poles = intraband.pole_residue.poles
            interband_poles = interband.pole_residue.poles
            poles = intraband_poles + interband_poles
            pole_residue = self._filter_poles(
                PoleResidue(poles=poles, frequency_range=(0, GRAPHENE_FIT_FREQ_MAX))
            )
            return Medium2D(ss=pole_residue, tt=pole_residue)
        return Medium2D(ss=intraband, tt=intraband)

    @property
    def intraband_drude(self) -> Drude:
        """A Drude-type model for the intraband term of graphene.

        Returns
        -------
        :class:`.Drude`
            A Drude-type model for the intraband term of graphene.

        """
        factor1 = Q_e * K_B * self.temp / (HBAR**2 * 4 * np.pi**3 * EPSILON_0)
        factor2 = self.mu_c / (K_B * self.temp) + 2 * np.log(
            np.exp(-self.mu_c / (K_B * self.temp)) + 1
        )
        intra_f1 = np.sqrt(self.scaling * factor1 * factor2)
        intra_delta1 = (1 / HBAR) * self.gamma / np.pi
        return Drude(coeffs=[(intra_f1, intra_delta1)])

    @property
    def interband_pole_residue(self) -> PoleResidue:
        """A pole-residue model for the interband term of graphene. Note that this does not
        include the intraband term, which is added in separately.

        Returns
        -------
        :class:`.PoleResidue`
            A pole-residue model for the interband term of graphene.
        """
        mu_c = self.mu_c / (2 * np.pi * HBAR)
        temp = self.temp * K_B / (2 * np.pi * HBAR)
        resonance = max(np.sqrt(abs(mu_c**2 - temp**2)), GRAPHENE_FIT_FREQ_MIN * 1e-5)
        freqs = np.linspace(
            resonance / GRAPHENE_FIT_LARGE_MULTIPLIER,
            resonance * GRAPHENE_FIT_LARGE_MULTIPLIER,
            GRAPHENE_FIT_NUM_FREQS,
        )
        sigma = self.interband_conductivity(freqs)
        nodes = self.interband_fit_freq_nodes
        if nodes is None:
            fcenter = freqs[np.argmin(np.imag(sigma))]
            fwidth = (
                fcenter - freqs[np.nonzero(np.imag(sigma) < 0.5 * np.amin(np.imag(sigma)))[0][0]]
            )
            nodes = [
                (fcenter / GRAPHENE_FIT_LARGE_MULTIPLIER, fcenter * GRAPHENE_FIT_LARGE_MULTIPLIER),
                (fcenter, fcenter + fwidth * GRAPHENE_FIT_SMALL_MULTIPLIER),
            ]

        flattened_freqs = [freq for pair in nodes for freq in pair]
        sigma_inds = self.interband_conductivity(flattened_freqs)
        inds = [(2 * i, 2 * i + 1) for i in range(len(nodes))]

        pole_residue = self._fit_interband_conductivity(flattened_freqs, sigma_inds, inds)
        pole_residue_filtered = self._filter_poles(pole_residue)
        # silence warnings associated with frequency range just for checking fit
        sigma_fit = pole_residue_filtered.updated_copy(frequency_range=None).sigma_model(freqs)
        if not np.allclose(sigma, sigma_fit, rtol=0, atol=GRAPHENE_FIT_ATOL):
            log.warning(
                "Graphene fit may not be good. Try changing the physical or fitting parameters."
            )
        return pole_residue_filtered

    def numerical_conductivity(self, freqs: List[float]) -> List[complex]:
        """Numerically calculate the conductivity. If this differs from the
        conductivity of the :class:`.Medium2D`, it is due to error while
        fitting the interband term, and you may try values of ``interband_fit_freq_nodes``
        different from its default (calculated) value.

        Parameters
        ----------
        freqs : List[float]
            The list of frequencies.

        Returns
        -------
        List[complex]
            The list of corresponding conductivities, in S.
        """
        intra_sigma = self.intraband_drude.sigma_model(freqs)
        inter_sigma = self.interband_conductivity(freqs)
        return intra_sigma + inter_sigma

    def interband_conductivity(self, freqs: List[float]) -> List[complex]:
        """Numerically integrate interband term.

        Parameters
        ----------
        freqs : List[float]
            The list of frequencies.

        Returns
        -------
        List[complex]
            The list of corresponding interband conductivities, in S.
        """

        def fermi(E: float) -> float:
            """Fermi distribution."""
            # catch overflow warning
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return 1 / (np.exp((E - self.mu_c) / (K_B * self.temp)) + 1)

        def fermi_g(E: float) -> float:
            """Difference of fermi distributions."""
            return fermi(-E) - fermi(E)

        def integrand(E: float, omega: float) -> float:
            """Integrand for interband term."""
            return (fermi_g(E * HBAR) - fermi_g(HBAR * omega / 2)) / (omega**2 - 4 * E**2)

        if not INTEGRATE_AVAILABLE:
            raise ImportError(
                "The package 'scipy' was not found. Please install the 'core' "
                "dependencies to calculate the interband term of graphene. For example: "
                "pip install tidy3d"
            )

        omegas = 2 * np.pi * np.array(freqs)
        sigma = np.zeros(len(omegas), dtype=complex)
        integration_min = GRAPHENE_INT_MIN
        integration_max = GRAPHENE_INT_MAX
        for i, omega in enumerate(omegas):
            integral, _ = integrate.quad(
                integrand, integration_min, integration_max, args=(omega,), epsabs=GRAPHENE_INT_TOL
            )
            sigma[i] = (Q_e / (4 * HBAR)) * (
                fermi_g(HBAR * omega / 2) - (4 * omega / (1j * np.pi)) * integral
            )

        return self.scaling * sigma

    def _fit_interband_conductivity(
        self,
        freqs: List[float],
        sigma: List[complex],
        indslist: List[Tuple[int, int]],
    ):
        """Fit the interband conductivity with a Pade approximation, as described in

        Stamatios Amanatiadis, Theodoros Zygiridis, Tadao Ohtani, Yasushi Kanai,
        and Nikolaos Kantartzis, "A Consistent Scheme for the Precise FDTD Modeling
        of the Graphene Interband Contribution," IEEE Trans. Magn. 57, 6 (2021).

        Parameters
        ----------
        freqs : List[float]
            The input frequencies.
        sigma : List[complex]
            The interband conductivity to fit.
        indslist : List[Tuple[int, int]]
            The indices at which to sample the data for fitting.
            The length of this list determines the number of Pade terms used.
        Returns
        -------
        :class:`.PoleResidue`
            A pole-residue model approximating the interband conductivity.
        """

        def evaluate_coeffslist(omega: List[float], coeffslist: List[List[float]]) -> List[float]:
            """Evaluate the Pade approximants given by ``coeffslist` to ``omega``.
            Each item in ``coeffslist`` is a list of four coefficients corresponding to
            a single Pade term."""
            res = np.zeros(len(omega), dtype=complex)
            for alpha0, alpha1, beta1, beta2 in coeffslist:
                res += (alpha0 + alpha1 * 1j * omega) / (
                    1 + beta1 * 1j * omega + beta2 * (1j * omega) ** 2
                )
            return res

        def fit_single(
            omega: List[float], sigma: List[complex], inds: Tuple[int, int]
        ) -> List[float]:
            """Fit a single Pade approximant of degree (1, 2) to ``sigma``
            as a real function of i ``omega``. The method is described in

            Adam Mock, "Pade approximant spectral fit for FDTD
            simulation of graphene in the near infrared,"
            Opt. Mater. Express 2, 6, pp. 771-781 (2012).
            """
            gamma = [np.real(sigma[i]) for i in inds]
            eta = [np.imag(sigma[i]) for i in inds]
            omegas = [omega[i] for i in inds]
            matrix = np.array(
                [
                    [1, 0, omegas[0] * eta[0], omegas[0] ** 2 * gamma[0]],
                    [0, omegas[0], -omegas[0] * gamma[0], omegas[0] ** 2 * eta[0]],
                    [1, 0, omegas[1] * eta[1], omegas[1] ** 2 * gamma[1]],
                    [0, omegas[1], -omegas[1] * gamma[1], omegas[1] ** 2 * eta[1]],
                ]
            )
            return np.linalg.pinv(matrix) @ np.array([gamma[0], eta[0], gamma[1], eta[1]])

        def optimize(
            omega: List[float],
            sigma: List[complex],
            indslist: List[Tuple[int, int]],
            coeffslist: List[List[float]],
        ) -> List[float]:
            """Optimize the coefficients in ``coeffslist`` by sampling ``omega`` and ``sigma``
            at the indices in ``indslist``."""
            for _ in range(self.interband_fit_num_iters):
                old_coeffslist = coeffslist
                res = sigma - evaluate_coeffslist(omega, old_coeffslist)
                for j, coeffs in enumerate(old_coeffslist):
                    curr_res = res + evaluate_coeffslist(omega, [coeffs])
                    coeffslist[j] = fit_single(omega, curr_res, indslist[j])
            return coeffslist

        def get_pole_residue(coeffslist: List[List[float]]) -> PoleResidue:
            """Convert a list of Pade coefficients into a :class:`.PoleResidue` model."""
            poles = []
            for alpha0, alpha1, beta1, beta2 in coeffslist:
                disc = beta1**2 - 4 * beta2
                root1 = (beta1 + np.sqrt(complex(disc))) / (2 * beta2)
                root2 = (beta1 - np.sqrt(complex(disc))) / (2 * beta2)
                res1 = (alpha0 - alpha1 * root1) / (beta2 * (root2 - root1))
                res2 = alpha1 / beta2 - res1

                if disc > 0:
                    for root, res in zip([root1, root2], [res1, res2]):
                        poles.append((root, res / 2))
                else:
                    poles.append((root1, res1))

            flipped_poles = []
            for a, c in poles:
                if np.real(a) > 0:
                    flipped_poles += [(-1j * np.conj(1j * a), c)]
                else:
                    flipped_poles += [(a, c)]
            return PoleResidue(poles=flipped_poles, frequency_range=(0, GRAPHENE_FIT_FREQ_MAX))

        # fitting works better with normalized quantities (THz and uS)
        omega_thz = 2 * np.pi * np.array(freqs) * 1e-12
        sigma_us = np.array(sigma) * 1e6
        coeffslist = []
        for inds in indslist:
            res = np.array(sigma_us) - evaluate_coeffslist(omega_thz, coeffslist)
            coeffslist.append(fit_single(omega_thz, res, inds))

        coeffslist = optimize(omega_thz, sigma_us, indslist, coeffslist)

        pole_res = get_pole_residue(coeffslist)
        # unnormalize, and convert from conductivity to permittivity
        poles = [(a * 1e12, -c / (a * EPSILON_0 * 1e6)) for (a, c) in pole_res.poles]

        return PoleResidue(poles=poles, frequency_range=(0, GRAPHENE_FIT_FREQ_MAX))

    def _filter_poles(self, medium: PoleResidue) -> PoleResidue:
        """Clean up poles, merging poles at zero frequency."""
        zero_res = 0
        poles = []
        for a, c in medium.poles:
            if a == 0:
                zero_res += c
            elif abs(a) > 1e17:
                continue
            else:
                poles += [(a, c)]
        return PoleResidue(
            poles=poles + [(0, zero_res)],
            frequency_range=(0, GRAPHENE_FIT_FREQ_MAX),
        )



"""Defines design space specification for tidy3d."""

from __future__ import annotations

import inspect
from typing import Any, Callable, Dict, List, Tuple, Union

import pydantic.v1 as pd

from ...components.base import TYPE_TAG_STR, Tidy3dBaseModel, cached_property
from ...components.data.sim_data import SimulationData
from ...components.simulation import Simulation
from ...log import Console, get_logging_console, log
from ...web.api.container import Batch, BatchData, Job
from .method import (
    MethodBayOpt,
    MethodGenAlg,
    MethodOptimize,
    MethodParticleSwarm,
    MethodType,
)
from .parameter import ParameterAny, ParameterInt, ParameterType
from .result import Result


class DesignSpace(Tidy3dBaseModel):
    """Manages all exploration of a parameter space within specified parameters using a supplied search method.
    The ``DesignSpace`` forms the basis of the ``Design`` plugin, and receives a ``Method`` and ``Parameter`` list that
    define the scope of the design space and how it should be searched. ``DesignSpace.run()`` can then be called with
    a function(s) to generate different solutions from parameters suggested by the ``Method``. The ``Method`` can either
    sample the design space systematically or randomly, or can optimize for a given problem through an iterative search
    and evaluate approach.


    Notes
    -----

        Schematic outline of how to use the ``Design`` plugin to explore a design space.

        .. image:: ../../../../_static/img/design.png
            :width: 80%
            :align: center

        The `Design <https://www.flexcompute.com/tidy3d/examples/notebooks/Design/>`_ notebook contains an overview of the
        ``Design`` plugin and is the best place to learn how to get started.
        Detailed examples using the ``Design`` plugin can be found in the following notebooks:

        * `All-Dielectric Structural Colors <https://www.flexcompute.com/tidy3d/examples/notebooks/AllDielectricStructuralColor/>`_
        * `Bayesian Optimization of Y-Junction <https://www.flexcompute.com/tidy3d/examples/notebooks/BayesianOptimizationYJunction/>`_
        * `Genetic Algorithm Reflector <https://www.flexcompute.com/tidy3d/examples/notebooks/GeneticAlgorithmReflector/>`_
        * `Particle Swarm Optimizer PBS <https://www.flexcompute.com/tidy3d/examples/notebooks/ParticleSwarmOptimizedPBS/>`_
        * `Particle Swarm Optimizer Bullseye Cavity <https://www.flexcompute.com/tidy3d/examples/notebooks/BullseyeCavityPSO/>`_

    Example
    -------
    >>> import tidy3d.plugins.design as tdd
    >>> param = tdd.ParameterFloat(name="x", span=(0,1))
    >>> method = tdd.MethodMonteCarlo(num_points=10)
    >>> design_space = tdd.DesignSpace(parameters=[param], method=method)
    >>> fn = lambda x: x**2
    >>> result = design_space.run(fn)
    >>> df = result.to_dataframe()
    >>> im = df.plot()

    """

    parameters: Tuple[ParameterType, ...] = pd.Field(
        (),
        title="Parameters",
        description="Set of parameters defining the dimensions and allowed values for the design space.",
    )

    method: MethodType = pd.Field(
        ...,
        title="Search Type",
        description="Specifications for the procedure used to explore the parameter space.",
        discriminator=TYPE_TAG_STR,  # Stops pydantic trying to validate every method whilst checking MethodType
    )

    task_name: str = pd.Field(
        "",
        title="Task Name",
        description="Task name assigned to tasks along with a simulation counter in the form of {task_name}_{sim_index}_{counter} where ``sim_index`` is "
        "the index of the ``Simulation`` from the pre function output. "
        "If the pre function outputs a dictionary the key will be included in the task name as {task_name}_{dict_key}_{counter}. "
        "Only used when pre-post functions are supplied.",
    )

    name: str = pd.Field(None, title="Name", description="Optional name for the design space.")

    path_dir: str = pd.Field(
        ".",
        title="Path Directory",
        description="Directory where simulation data files will be locally saved to. Only used when pre and post functions are supplied.",
    )

    folder_name: str = pd.Field(
        "default",
        title="Folder Name",
        description="Folder path where the simulation will be uploaded in the Tidy3D Workspace. Will use 'default' if no path is set.",
    )

    @cached_property
    def dims(self) -> Tuple[str]:
        """dimensions defined by the design parameter names."""
        return tuple(param.name for param in self.parameters)

    def _package_run_results(
        self,
        fn_args: list[dict[str, Any]],
        fn_values: List[Any],
        fn_source: str,
        task_names: Tuple[str] = None,
        task_paths: list = None,
        aux_values: List[Any] = None,
        opt_output: Any = None,
    ) -> Result:
        """How to package results from ``method.run`` and ``method.run_batch``"""

        fn_args_coords = tuple(
            [[arg_dict[key] for arg_dict in fn_args] for key in fn_args[0].keys()]
        )

        fn_args_coords_T = list(map(list, zip(*fn_args_coords)))

        return Result(
            dims=tuple(fn_args[0].keys()),
            values=fn_values,
            coords=fn_args_coords_T,
            fn_source=fn_source,
            task_names=task_names,
            task_paths=task_paths,
            aux_values=aux_values,
            optimizer=opt_output,
        )

    @staticmethod
    def get_fn_source(function: Callable) -> str:
        """Get the function source as a string, return ``None`` if not available."""
        try:
            return inspect.getsource(function)
        except (TypeError, OSError):
            return None

    def run(self, fn: Callable, fn_post: Callable = None, verbose: bool = True) -> Result:
        """Explore a parameter space with a supplied method using the user supplied function.
        Supplied functions are used to evaluate the design space and are called within the method.
        For optimization methods these functions act as the fitness function. A single function can be
        supplied which will contain the preprocessing, computation, and analysis of the desired problem.
        If running a Tidy3D simulation, it is recommended to split this function into a pre function, that creates a Simulation object(s),
        and a post function which analyses the SimulationData produced by the pre function Simulations. This allows the DesignSpace to
        manage the batching of Simulations, which varies between Method used, and saving time writing their own batching code.
        It also efficiently submits simulations to the cloud servers allowing for the fastest exploration of a design space.

        The ``fn`` function must take a dictionary input - this can be stored a dictionary ``def example_fn(**params)``
        or left as keyword arguments ``def example_fn(arg1, arg2)`` where the keywords correspond to the ``name`` of the parameters in the design space.

        If used as a pre function, the output of ``fn`` must be a float, ``Simulation``, ``Batch``, list, or dict. Supplied ``Batch`` objects are
        run without modification and are run in series. A list or dict of ``Simulation`` objects is flattened into a single ``Batch`` to enable
        parallel computation on the cloud. The original structure is then restored for output; all ``Simulation`` objects are replaced by ``SimulationData`` objects.
        Example pre return formats and associated post inputs can be seen in the table below.

        .. list-table:: Pre return formats and post input formats
            :widths: 50 50
            :header-rows: 1

            * - fn_pre return
              - fn_post call
            * - 1.0
              - fn_post(1.0)
            * - [1,2,3]
              - fn_post(1,2,3)
            * - {'a': 2, 'b': 'hi'}
              - fn_post(a=2, b='hi')
            * - Simulation
              - fn_post(SimulationData)
            * - Batch
              - fn_post(BatchData)
            * - [Simulation, Simulation]
              - fn_post(SimulationData, SimulationData)
            * - [Simulation, 1.0]
              - fn_post(SimulationData, 1.0)
            * - [Simulation, Batch]
              - fn_post(SimulationData, BatchData)
            * - {'a': Simulation, 'b': Batch, 'c': 2.0}
              - fn_post(a=SimulationData, b=BatchData, c=2.0)

        The output of ``fn_post`` (or ``fn`` if only one function is supplied) must be a float
        or a container where the first element is a ``float`` and second element is a ``list`` / ``dict`` e,g. [float {"aux_1": str}].
        The float is used by the optimizers as the return of the fitness function.
        The second element is for auxiliary data from the analysis that the user may want to keep.
        Sampling methods (``MethodGrid`` or ``MethodMonteCarlo``) can have any return type.

        Parameters
        ----------
        fn : Callable
            Function accepting arguments that correspond to the ``name`` fields
            of the ``DesignSpace.parameters``.
            Must return in the expected format for the ``method`` used in ``DesignSpace``,
            or return an object that fn_post can accept as an input.

        fn_post : Callable = None
            Optional function performing postprocessing on the output of ``fn``.
            It is recommended to supply fn_post when working with Simulation objects.
            Must return in the expected format for the ``method`` used in ``DesignSpace``.

        verbose : bool = True
            Toggle the output of statements stored in the logging console.

        Returns
        -------
        :class:`Result`
            Object containing the results of the design space exploration.
            Can be converted to ``pandas.DataFrame`` with ``.to_dataframe()``.
        """

        # Get the console
        # Method.run checks for console is None instead of being passed console and verbose
        console = get_logging_console() if verbose else None

        # Run based on how many functions the user provides
        if fn_post is None:
            fn_args, fn_values, aux_values, opt_output = self.run_single(fn, console)
            sim_names = None
            sim_paths = None

        else:
            fn_args, fn_values, aux_values, opt_output, sim_names, sim_paths = self.run_pre_post(
                fn_pre=fn, fn_post=fn_post, console=console
            )

            if len(sim_names) == 0:
                sim_names = None

        fn_source = self.get_fn_source(fn)

        # Package the result
        return self._package_run_results(
            fn_args=fn_args,
            fn_values=fn_values,
            fn_source=fn_source,
            aux_values=aux_values,
            task_names=sim_names,
            task_paths=sim_paths,
            opt_output=opt_output,
        )

    def run_single(self, fn: Callable, console: Console) -> Tuple(list[dict], list, list[Any]):
        """Run a single function of parameter inputs."""
        evaluate_fn = self._get_evaluate_fn_single(fn=fn)
        return self.method._run(run_fn=evaluate_fn, parameters=self.parameters, console=console)

    def run_pre_post(self, fn_pre: Callable, fn_post: Callable, console: Console) -> Tuple(
        list[dict], list[dict], list[Any]
    ):
        """Run a function with Tidy3D implicitly called in between."""
        handler = self._get_evaluate_fn_pre_post(
            fn_pre=fn_pre, fn_post=fn_post, fn_mid=self._fn_mid, console=console
        )
        fn_args, fn_values, aux_values, opt_output = self.method._run(
            run_fn=handler.fn_combined, parameters=self.parameters, console=console
        )
        return fn_args, fn_values, aux_values, opt_output, handler.sim_names, handler.sim_paths

    """ Helpers """

    def _get_evaluate_fn_single(self, fn: Callable) -> list[Any]:
        """Get function that sequentially evaluates single `fn` for a list of arguments."""

        def evaluate(args_list: list) -> list[Any]:
            """Evaluate a list of arguments passed to ``fn``."""
            return [fn(**args) for args in args_list]

        return evaluate

    def _get_evaluate_fn_pre_post(
        self, fn_pre: Callable, fn_post: Callable, fn_mid: Callable, console: Console
    ):
        """Get function that tries to use batch processing on a set of arguments."""

        class Pre_Post_Handler:
            def __init__(self, console):
                self.sim_counter = 0
                self.sim_names = []
                self.sim_paths = []
                self.console = console

            def fn_combined(self, args_list: list[dict[str, Any]]) -> list[Any]:
                """Compute fn_pre and fn_post functions and capture other outputs."""
                sim_dict = {str(idx): fn_pre(**arg_list) for idx, arg_list in enumerate(args_list)}

                if not all(
                    isinstance(val, (int, float, Simulation, Batch, list, dict))
                    for val in sim_dict.values()
                ):
                    raise ValueError(
                        "Unrecognized output of fn_pre. Please change the return of fn_pre."
                    )

                data, task_names, task_paths, sim_counter = fn_mid(
                    sim_dict, self.sim_counter, self.console
                )
                self.sim_names.extend(task_names)
                self.sim_paths.extend(task_paths)
                self.sim_counter = sim_counter
                post_out = [fn_post(val) for val in data.values()]
                return post_out

        handler = Pre_Post_Handler(console)

        return handler

    def _fn_mid(
        self, pre_out: dict[int, Any], sim_counter: int, console: Console
    ) -> Union[dict[int, Any], BatchData]:
        """A function of the output of ``fn_pre`` that gives the input to ``fn_post``."""

        # Keep copy of original to use if no tidy3d simulation required
        original_pre_out = pre_out.copy()

        # Convert list input to dict of dict before passing through checks
        was_list = False
        if all(isinstance(sim, list) for sim in pre_out.values()):
            pre_out = {
                str(idx): {str(sub_idx): sub_list for sub_idx, sub_list in enumerate(sim_list)}
                for idx, sim_list in enumerate(pre_out.values())
            }
            was_list = True

        def _find_and_map(
            search_dict: dict,
            search_type: Any,
            output_dict: dict,
            naming_dict: dict,
            previous_key: str = "",
        ):
            """Recursively search for search_type objects within a dictionary."""
            current_key = previous_key
            for key, value in search_dict.items():
                if not len(previous_key):
                    latest_key = str(key)
                else:
                    latest_key = f"{current_key}_{key}"

                if isinstance(value, dict):
                    _find_and_map(value, search_type, output_dict, naming_dict, latest_key)
                elif isinstance(value, search_type):
                    output_dict[latest_key] = value
                    naming_dict[latest_key] = key

        simulations = {}
        batches = {}
        naming_keys = {}

        _find_and_map(pre_out, Simulation, simulations, naming_keys)
        _find_and_map(pre_out, Batch, batches, naming_keys)

        # Exit fn_mid here if no td computation is required
        if not len(simulations) and not len(batches):
            return original_pre_out, list(), list(), sim_counter

        # Create task names for simulations
        named_sims = {}
        translate_sims = {}
        for sim_key, sim in simulations.items():
            # Checks stop standard indexing keys being included in the name
            suffix = f"{naming_keys[sim_key]}_{sim_counter}"

            # Handle if the user does not want a task name
            if len(self.task_name) > 0:
                sim_name = f"{self.task_name}_{suffix}"
            else:
                sim_name = suffix
            named_sims[sim_name] = sim
            translate_sims[sim_name] = sim_key
            sim_counter += 1

        # Log the simulations and batches for the user
        if console is not None:
            # Writen like this to include batches on the same line if present
            run_statement = f"{len(named_sims)} Simulations"
            if len(batches) > 0:
                run_statement = run_statement + f" and {len(batches)} user Batches"

            console.log(f"Running {run_statement}")

        # Running simulations and batches
        sims_out = Batch(
            simulations=named_sims,
            folder_name=self.folder_name,
            simulation_type="tidy3d_design",
            verbose=False,  # Using a custom output instead of Batch.monitor updates
        ).run(path_dir=self.path_dir)

        batch_results = {}
        for batch_key, batch in batches.items():
            batch_out = batch.run(path_dir=self.path_dir)
            batch_results[batch_key] = batch_out

        def _return_to_dict(return_dict: dict, key: str, return_obj: Any) -> None:
            """Recursively insert items into a dict by keys split with underscore. Only works for dict or dict of dict inputs."""
            split_key = key.split("_", 1)
            if len(split_key) > 1:
                _return_to_dict(return_dict[split_key[0]], split_key[1], return_obj)
            else:
                return_dict[split_key[0]] = return_obj

        for (sim_name, sim), task_name, task_path in zip(
            sims_out.items(), sims_out.task_ids.keys(), sims_out.task_paths.values()
        ):
            translated_name = translate_sims[sim_name]
            sim.attrs["task_name"] = task_name
            sim.attrs["task_path"] = task_path
            _return_to_dict(pre_out, translated_name, sim)

        for batch_name, batch in batch_results.items():
            _return_to_dict(pre_out, batch_name, batch)

        def _remove_or_replace(search_dict: dict, attr_name: str) -> dict:
            """Recursively search through a dict replacing Sims and Batches or ignoring other items thus removing them"""
            new_dict = {}
            for key, value in search_dict.items():
                if isinstance(value, dict):
                    new_sub_dict = _remove_or_replace(value, attr_name)
                    new_dict[key] = new_sub_dict

                else:
                    if isinstance(value, SimulationData):
                        new_dict[key] = value.attrs[attr_name]

                    elif isinstance(value, BatchData):
                        if attr_name == "task_name":
                            new_dict[key] = list(value.task_ids.keys())
                        else:
                            new_dict[key] = list(value.task_paths.values())

            return new_dict

        # Build out a dict of task_name or task_path in the same shape as the original data
        task_names = _remove_or_replace(pre_out.copy(), "task_name")
        task_paths = _remove_or_replace(pre_out.copy(), "task_path")

        # Reduce down to a list to be extended later
        task_names = list(task_names.values())
        task_paths = list(task_paths.values())

        # Restore output to a list if a list was supplied
        if was_list:
            pre_out = {dict_idx: list(sub_dict.values()) for dict_idx, sub_dict in pre_out.items()}
            task_names = [list(sub_dict.values()) for sub_dict in task_names]
            task_paths = [list(sub_dict.values()) for sub_dict in task_paths]

        return pre_out, task_names, task_paths, sim_counter

    def run_batch(
        self,
        fn_pre: Callable[Any, Union[Simulation, List[Simulation], Dict[str, Simulation]]],
        fn_post: Callable[
            Union[SimulationData, List[SimulationData], Dict[str, SimulationData]], Any
        ],
        path_dir: str = ".",
        **batch_kwargs,
    ) -> Result:
        """
        This function has been superceded by `run`, please use `run` for batched simulations.
        """
        log.warning(
            "In version 2.8.0, the 'run_batch' method is replaced by 'run'."
            "'fn_pre' has become 'fn', whilst 'fn_post' remains the same."
            "While the original syntax will still be supported, future functionality will be added to the new method moving forward."
        )

        if len(batch_kwargs) > 0:
            log.warning(
                "'batch_kwargs' supplied here will no longer be used within the simulation."
            )

        new_self = self.updated_copy(path_dir=path_dir)
        result = new_self.run(fn=fn_pre, fn_post=fn_post)

        return result

    def estimate_cost(self, fn_pre: Callable) -> float:
        """Compute the maximum FlexCredit charge for the ``DesignSpace.run`` computation.

        Require a pre function that should return a ``Simulation`` object, a ``Batch`` object, or collection of either.
        The pre function is called to estimate the cost - complicated pre functions may cause long runtimes. The cost per
        iteration is multiplied by the theoretical maximum number of iterations to give the maximum cost.

        Parameters
        ----------
        fn_pre : Callable
            Function accepting arguments that correspond to the ``name`` fields
            of the ``DesignSpace.parameters``. Should return a ``Simulation`` or ``Batch`` object, or a
            ``list`` / ``dict`` of these objects.

        Returns
        -------
        float
            Estimated maximum cost for the ``DesignSpace.run``.

        """
        # Get output fn_pre for paramters at the lowest span / default
        arg_dict = {}
        for param in self.parameters:
            arg_dict[param.name] = param.sample_first()

        # Compute fn_pre
        pre_out = fn_pre(**arg_dict)

        def _estimate_sim_cost(sim):
            job = Job(simulation=sim, task_name="estimate_cost")

            estimate = job.estimate_cost()
            job.delete()  # Deleted as only a test with initial parameters

            return estimate

        if isinstance(pre_out, Simulation):
            per_run_estimate = _estimate_sim_cost(pre_out)

        elif isinstance(pre_out, Batch):
            per_run_estimate = pre_out.estimate_cost()
            pre_out.delete()  # Deleted as only a test with initial parameters

        elif isinstance(pre_out, (list, dict)):
            # Iterate through container to get simulations and batches and sum cost
            # Accept list or dict inputs

            if isinstance(pre_out, dict):
                pre_out = list(pre_out.values())

            sims = []
            batches = []
            for value in pre_out:
                if isinstance(value, Simulation):
                    sims.append(value)
                elif isinstance(value, Batch):
                    batches.append(value)

            calculated_estimates = []
            for sim in sims:
                calculated_estimates.append(_estimate_sim_cost(sim))

            for batch in batches:
                calculated_estimates.append(batch.estimate_cost())
                batch.delete()  # Deleted as only a test with initial parameters

            if None in calculated_estimates:
                per_run_estimate = None
            else:
                per_run_estimate = sum(calculated_estimates)

        else:
            raise ValueError("Unrecognized output from pre-function, unable to estimate cost.")

        # Calculate maximum number of runs for different methods
        run_count = self.method._get_run_count(self.parameters)

        # For if tidy3d server cannot determine the estimate
        if per_run_estimate is None:
            return None
        else:
            return round(per_run_estimate * run_count, 3)

    def summarize(self, fn_pre: Callable = None, verbose: bool = True) -> dict[str, Any]:
        """Summarize the setup of the DesignSpace

        Prints a summary of the DesignSpace including the method and associated args, the parameters,
        and the maximum number of runs expected. If ``fn_pre`` is provided an estimated cost will
        also be included. Additional notes are printed where relevant. All data is returned as a dict.

        Parameters
        ----------
        fn_pre : Callable = None
            Function accepting arguments that correspond to the ``name`` fields
            of the ``DesignSpace.parameters``. Allows for estimated cost to be included
            in the summary.
        verbose: bool = True
            Toggle if the summary should be output to log. If False, the dict is returned silently.

        Returns
        -------
        summary_dict: dict
            Dictionary containing the summary information.
        """

        # Get output console
        console = get_logging_console()

        # Assemble message
        # If check stops it printing standard attributes
        arg_values = [
            f"{field}: {getattr(self.method, field)}\n"
            for field in self.method.__fields__
            if field not in MethodOptimize.__fields__
        ]

        param_values = []
        for param in self.parameters:
            if isinstance(param, ParameterAny):
                param_values.append(f"{param.name}: {param.type} {param.allowed_values}\n")
            else:
                param_values.append(f"{param.name}: {param.type} {param.span}\n")

        run_count = self.method._get_run_count(self.parameters)

        # Compile data into a dict for return
        summary_dict = {
            "method": self.method.type,
            "method_args": "".join(arg_values),
            "param_count": len(self.parameters),
            "param_names": ", ".join([param.name for param in self.parameters]),
            "param_vals": "".join(param_values),
            "max_run_count": run_count,
        }

        if verbose:
            console.log(
                "\nSummary of DesignSpace\n\n"
                f"Method: {summary_dict['method']}\n"
                f"Method Args\n{summary_dict['method_args']}\n"
                f"No. of Parameters: {summary_dict['param_count']}\n"
                f"Parameters: {summary_dict['param_names']}\n"
                f"{summary_dict['param_vals']}\n"
                f"Maximum Run Count: {summary_dict['max_run_count']}\n"
            )

            if fn_pre is not None:
                cost_estimate = self.estimate_cost(fn_pre)
                summary_dict["cost_estimate"] = cost_estimate
                console.log(f"Estimated Maximum Cost: {cost_estimate} FlexCredits")

                # NOTE: Could then add more details regarding the output of fn_pre - confirm batching?

        # Include additional notes/warnings
        notes = []

        if isinstance(self.method, MethodGenAlg):
            notes.append(
                "The maximum run count for MethodGenAlg is difficult to predict. "
                "Repeated solutions are not executed, reducing the total number of simulations. "
                "High crossover and mutation probabilities may result in an increased number of simulations, potentially exceeding the predicted maximum run count."
            )

        if isinstance(self.method, (MethodBayOpt, MethodGenAlg, MethodParticleSwarm)):
            if any(isinstance(param, ParameterInt) for param in self.parameters):
                if any(isinstance(param, ParameterAny) for param in self.parameters):
                    notes.append(
                        "Discrete 'ParameterAny' values are automatically converted to 'int' values to be optimized.\n"
                    )

                notes.append(
                    "Discrete 'int' values are automatically rounded if optimizers generate 'float' predictions.\n"
                )

        if len(notes) > 0 and verbose:
            console.log(
                "Notes:",
            )
            for note in notes:
                console.log(note)

        return summary_dict



"""Fit PoleResidue Dispersion models to optical NK data"""

from __future__ import annotations

from typing import Tuple

import numpy as np
from pydantic.v1 import NonNegativeFloat, PositiveInt

from ...components.dispersion_fitter import AdvancedFastFitterParam, fit
from ...components.medium import PoleResidue
from ...constants import C_0, HBAR
from .fit import DispersionFitter

# numerical tolerance for pole relocation for fast fitter
TOL = 1e-8
# numerical cutoff for passivity testing
CUTOFF = np.finfo(np.float32).eps
# parameters for passivity optimization
PASSIVITY_NUM_ITERS_DEFAULT = 50
SLSQP_CONSTRAINT_SCALE_DEFAULT = 1e35
# min value of the rms for default weights calculated based on Re[eps], Im[eps]
RMS_MIN = 0.1

DEFAULT_MAX_POLES = 5
DEFAULT_NUM_ITERS = 20
DEFAULT_TOLERANCE_RMS = 1e-5

# this avoids divide by zero errors with lossless poles
SCALE_FACTOR = 1.01

# when poles are close to omega, can cause invalid response function, and we reject model
OMEGA_POLE_CLOSE_ATOL = 1e-10


class FastDispersionFitter(DispersionFitter):
    """Tool for fitting refractive index data to get a
    dispersive medium described by :class:`.PoleResidue` model."""

    def fit(
        self,
        min_num_poles: PositiveInt = 1,
        max_num_poles: PositiveInt = DEFAULT_MAX_POLES,
        eps_inf: float = None,
        tolerance_rms: NonNegativeFloat = DEFAULT_TOLERANCE_RMS,
        advanced_param: AdvancedFastFitterParam = None,
    ) -> Tuple[PoleResidue, float]:
        """Fit data using a fast fitting algorithm.

        Note
        ----
        The algorithm is described in::

            B. Gustavsen and A. Semlyen, "Rational approximation
            of frequency domain responses by vector fitting,"
            IEEE Trans. Power. Deliv. 14, 3 (1999).

            B. Gustavsen, "Improving the pole relocation properties
            of vector fitting," IEEE Trans. Power Deliv. 21, 3 (2006).

            B. Gustavsen, "Enforcing Passivity for Admittance Matrices
            Approximated by Rational Functions," IEEE Trans. Power
            Syst. 16, 1 (2001).

        Note
        ----
        The fit is performed after weighting the real and imaginary parts,
        so the RMS error is also weighted accordingly. By default, the weights
        are chosen based on typical values of the data. To change this behavior,
        use 'AdvancedFastFitterParam.weights'.


        Parameters
        ----------
        min_num_poles: PositiveInt, optional
            Minimum number of poles in the model.
        max_num_poles: PositiveInt, optional
            Maximum number of poles in the model.
        eps_inf : float, optional
            Value of eps_inf to use in fit. If None, then eps_inf is also fit.
            Note: fitting eps_inf is not guaranteed to yield a global optimum, so
            the result may occasionally be better with a fixed value of eps_inf.
        tolerance_rms : float, optional
            Weighted RMS error below which the fit is successful and the result is returned.
        advanced_param : :class:`AdvancedFastFitterParam`, optional
            Advanced parameters for fitting.

        Returns
        -------
        Tuple[:class:`.PoleResidue`, float]
            Best fitting result: (dispersive medium, weighted RMS error).
        """

        omega_data = PoleResidue.Hz_to_angular_freq(self.freqs)
        eps_data = self.eps_data
        scale_factor = HBAR

        params, error = fit(
            omega_data=omega_data,
            resp_data=eps_data,
            min_num_poles=min_num_poles,
            max_num_poles=max_num_poles,
            resp_inf=eps_inf,
            tolerance_rms=tolerance_rms,
            advanced_param=advanced_param,
            scale_factor=scale_factor,
        )

        eps_inf, poles, residues = params

        medium = PoleResidue(eps_inf=eps_inf, poles=list(zip(poles, residues)))

        return medium, error

    @classmethod
    def constant_loss_tangent_model(
        cls,
        eps_real: float,
        loss_tangent: float,
        frequency_range: Tuple[float, float],
        max_num_poles: PositiveInt = DEFAULT_MAX_POLES,
        number_sampling_frequency: PositiveInt = 10,
        tolerance_rms: NonNegativeFloat = DEFAULT_TOLERANCE_RMS,
    ) -> PoleResidue:
        """Fit a constant loss tangent material model.

        Parameters
        ----------
        eps_real : float
            Real part of permittivity
        loss_tangent : float
            Loss tangent.
        frequency_range : Tuple[float, float]
            Freqquency range for the material to exhibit constant loss tangent response.
        max_num_poles : PositiveInt, optional
            Maximum number of poles in the model.
        number_sampling_frequency : PositiveInt, optional
            Number of sampling frequencies to compute RMS error for fitting.
        tolerance_rms : float, optional
            Weighted RMS error below which the fit is successful and the result is returned.

        Returns
        -------
        :class:`.PoleResidue
            Best results of multiple fits.
        """
        if number_sampling_frequency < 2:
            frequencies = np.array([np.mean(frequency_range)])
        else:
            frequencies = np.linspace(
                frequency_range[0], frequency_range[1], number_sampling_frequency
            )
        wvl_um = C_0 / frequencies
        eps_real_array = np.ones_like(frequencies) * eps_real
        loss_tangent_array = np.ones_like(frequencies) * loss_tangent
        fitter = cls.from_loss_tangent(wvl_um, eps_real_array, loss_tangent_array)
        material, _ = fitter.fit(max_num_poles=max_num_poles, tolerance_rms=tolerance_rms)
        return material



"""Defines electric current sources for injecting light into simulation."""

from __future__ import annotations

from typing import Union

from .current import CustomCurrentSource, PointDipole, UniformCurrentSource
from .field import (
    TFSF,
    AstigmaticGaussianBeam,
    CustomFieldSource,
    GaussianBeam,
    ModeSource,
    PlaneWave,
)

# sources allowed in Simulation.sources
SourceType = Union[
    UniformCurrentSource,
    PointDipole,
    GaussianBeam,
    AstigmaticGaussianBeam,
    ModeSource,
    PlaneWave,
    CustomFieldSource,
    CustomCurrentSource,
    TFSF,
]



"""Utilities for plotting microwave components"""

from numpy import inf

from ...components.viz import PathPlotParams

""" Constants """
VOLTAGE_COLOR = "red"
CURRENT_COLOR = "blue"
LOBE_PEAK_COLOR = "tab:red"
LOBE_WIDTH_COLOR = "tab:orange"
LOBE_FNBW_COLOR = "tab:blue"
PATH_LINEWIDTH = 2
ARROW_CURRENT = dict(
    arrowstyle="-|>",
    mutation_scale=32,
    linestyle="",
    lw=PATH_LINEWIDTH,
    color=CURRENT_COLOR,
)

plot_params_voltage_path = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=VOLTAGE_COLOR,
    linestyle="--",
    linewidth=PATH_LINEWIDTH,
    marker="o",
    markersize=10,
    markeredgecolor=VOLTAGE_COLOR,
    markerfacecolor="white",
)

plot_params_voltage_plus = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=VOLTAGE_COLOR,
    marker="+",
    markersize=6,
)

plot_params_voltage_minus = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=VOLTAGE_COLOR,
    marker="_",
    markersize=6,
)

plot_params_current_path = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=CURRENT_COLOR,
    linestyle="--",
    linewidth=PATH_LINEWIDTH,
    marker="",
)

plot_params_lobe_peak = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=LOBE_PEAK_COLOR,
    linestyle="-",
    linewidth=1,
    marker="",
)

plot_params_lobe_width = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=LOBE_WIDTH_COLOR,
    linestyle="--",
    linewidth=1,
    marker="",
)

plot_params_lobe_FNBW = PathPlotParams(
    alpha=1.0,
    zorder=inf,
    color=LOBE_FNBW_COLOR,
    linestyle=":",
    linewidth=1,
    marker="",
)



# ruff: noqa: W293, W291
"""Defines heat simulation class"""

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Tuple, Union

import numpy as np
import pydantic.v1 as pd

try:
    from matplotlib import colormaps
except ImportError:
    pass

from tidy3d.components.base import skip_if_fields_missing
from tidy3d.components.base_sim.simulation import AbstractSimulation
from tidy3d.components.bc_placement import (
    MediumMediumInterface,
    SimulationBoundary,
    StructureBoundary,
    StructureSimulationBoundary,
    StructureStructureInterface,
)
from tidy3d.components.geometry.base import Box
from tidy3d.components.material.tcad.charge import (
    ChargeConductorMedium,
    SemiconductorMedium,
)
from tidy3d.components.material.tcad.heat import (
    SolidMedium,
)
from tidy3d.components.material.types import MultiPhysicsMedium, StructureMediumType
from tidy3d.components.medium import Medium
from tidy3d.components.scene import Scene
from tidy3d.components.spice.sources.dc import DCVoltageSource
from tidy3d.components.spice.types import ElectricalAnalysisType
from tidy3d.components.structure import Structure
from tidy3d.components.tcad.boundary.specification import (
    HeatBoundarySpec,
    HeatChargeBoundarySpec,
)
from tidy3d.components.tcad.grid import (
    DistanceUnstructuredGrid,
    UniformUnstructuredGrid,
    UnstructuredGridType,
)
from tidy3d.components.tcad.monitors.charge import (
    SteadyCapacitanceMonitor,
    SteadyFreeCarrierMonitor,
    SteadyPotentialMonitor,
)
from tidy3d.components.tcad.monitors.heat import (
    TemperatureMonitor,
)
from tidy3d.components.tcad.source.abstract import (
    GlobalHeatChargeSource,
)
from tidy3d.components.tcad.types import (
    ConvectionBC,
    CurrentBC,
    HeatChargeMonitorType,
    HeatChargeSourceType,
    HeatFluxBC,
    HeatFromElectricSource,
    HeatSource,
    InsulatingBC,
    TemperatureBC,
    UniformHeatSource,
    VoltageBC,
)
from tidy3d.components.tcad.viz import (
    CHARGE_BC_INSULATOR,
    HEAT_BC_COLOR_CONVECTION,
    HEAT_BC_COLOR_FLUX,
    HEAT_BC_COLOR_TEMPERATURE,
    HEAT_SOURCE_CMAP,
    plot_params_heat_bc,
    plot_params_heat_source,
)
from tidy3d.components.types import TYPE_TAG_STR, Ax, Bound, ScalarSymmetry, Shapely, annotate_type
from tidy3d.components.viz import PlotParams, add_ax_if_none, equal_aspect
from tidy3d.constants import VOLUMETRIC_HEAT_RATE, inf
from tidy3d.exceptions import SetupError
from tidy3d.log import log

HEAT_CHARGE_BACK_STRUCTURE_STR = "<<<HEAT_CHARGE_BACKGROUND_STRUCTURE>>>"

HeatBCTypes = (TemperatureBC, HeatFluxBC, ConvectionBC)
HeatSourceTypes = (UniformHeatSource, HeatSource, HeatFromElectricSource)
ChargeSourceTypes = ()
ElectricBCTypes = (VoltageBC, CurrentBC, InsulatingBC)

AnalysisSpecType = ElectricalAnalysisType


class TCADAnalysisTypes(str, Enum):
    """Enumeration of the types of simulations currently supported"""

    HEAT = "Heat"
    CONDUCTION = "Conduction"
    CHARGE = "Charge"


class HeatChargeSimulation(AbstractSimulation):
    """
    Defines thermoelectric simulations.

    Notes
    -----
        A ``HeatChargeSimulation`` supports different types of simulations. It solves the
        heat and conduction equations using the Finite-Volume (FV) method. This solver
        determines the required computation physics according to the simulation scene definition.
        This is implemented in this way due to the strong multi-physics coupling.

    The ``HeatChargeSimulation`` can solve multiple physics and the intention is to enable close thermo-electrical coupling.

    Currently, this solver supports steady-state heat conduction where :math:`q` is the heat flux, :math:`k`
    is the thermal conductivity, and :math:`T` is the temperature.

         .. math::

            -k \\cdot \\nabla(T) = q


    The steady-state electrical ``Conduction`` equation depends on the electric conductivity (:math:`\\sigma`)  of a
    medium, and the electric field (:math:`\\mathbf{E} = -\\nabla(\\psi)`) derived from electrical potential (:math:`\\psi`).
    Currently, in this type of simulation, no current sources or sinks are supported.

        .. math::

            \\text{div}(\\sigma \\cdot \\nabla(\\psi)) = 0


    For further details on what equations are solved in ``Charge`` simulations, refer to the :class:`SemiconductorMedium`.

    Let's understand how the physics solving is determined:

        .. list-table::
           :widths: 25 75
           :header-rows: 1

           * - Simulation Type
             - Example Configuration Settings
           * - ``Heat``
             - The heat equation is solved with specified heat sources,
               boundary conditions, etc. Structures should incorporate materials
               with defined heat properties.
           * - ``Conduction``
             - The electrical conduction equation is solved with
               specified boundary conditions such as ``SteadyVoltageBC``, ``SteadyCurrentBC``, ...
           * - ``Charge``
             - Drift-diffusion equations are solved for structures containing
               a defined :class:`SemiconductorMedium`. Insulators with a
               :class:`ChargeInsulatorMedium` can also be included. For these, only the
               electric potential field is calculated.

    Examples
    --------
    To run a thermal (``Heat`` |:fire:|) simulation with a solid conductive structure:

    >>> import tidy3d as td
    >>> heat_sim = td.HeatChargeSimulation(
    ...     size=(3.0, 3.0, 3.0),
    ...     structures=[
    ...         td.Structure(
    ...             geometry=td.Box(size=(1, 1, 1), center=(0, 0, 0)),
    ...             medium=td.Medium(
    ...                 permittivity=2.0,
    ...                 heat_spec=td.SolidSpec(
    ...                     conductivity=1,
    ...                     capacity=1,
    ...                 )
    ...             ),
    ...             name="box",
    ...         ),
    ...     ],
    ...     medium=td.Medium(permittivity=3.0, heat_spec=td.FluidSpec()),
    ...     grid_spec=td.UniformUnstructuredGrid(dl=0.1),
    ...     sources=[td.HeatSource(rate=1, structures=["box"])],
    ...     boundary_spec=[
    ...         td.HeatChargeBoundarySpec(
    ...             placement=td.StructureBoundary(structure="box"),
    ...             condition=td.TemperatureBC(temperature=500),
    ...         )
    ...     ],
    ...     monitors=[td.TemperatureMonitor(size=(1, 2, 3), name="sample")],
    ... )

    To run a drift-diffusion (``Charge`` |:zap:|) system:
    >>> import tidy3d as td
    >>> air = td.FluidMedium(
    ...     name="air"
    ... )
    >>> intrinsic_Si = td.material_library['cSi'].variants['Si_MultiPhysics'].medium.charge
    >>> Si_n = intrinsic_Si.updated_copy(N_d=1e16, name="Si_n")
    >>> Si_p = intrinsic_Si.updated_copy(N_a=1e16, name="Si_p")
    >>> n_side = td.Structure(
    ...     geometry=td.Box(center=(-0.5, 0, 0), size=(1, 1, 1)),
    ...     medium=Si_n,
    ...     name="n_side"
    ... )
    >>> p_side = td.Structure(
    ...     geometry=td.Box(center=(0.5, 0, 0), size=(1, 1, 1)),
    ...     medium=Si_p,
    ...     name="p_side"
    ... )
    >>> bc_v1 = td.HeatChargeBoundarySpec(
    ...     condition=td.VoltageBC(source=td.DCVoltageSource(voltage=[-1, 0, 0.5])),
    ...     placement=td.MediumMediumInterface(mediums=[air.name, Si_n.name]),
    ... )
    >>> bc_v2 = td.HeatChargeBoundarySpec(
    ...     condition=td.VoltageBC(source=td.DCVoltageSource(voltage=0)),
    ...     placement=td.MediumMediumInterface(mediums=[air.name, Si_p.name]),
    ... )
    >>> charge_sim = td.HeatChargeSimulation(
    ...     structures=[n_side, p_side],
    ...     medium=td.Medium(heat_spec=td.FluidSpec(), name="air"),
    ...     monitors=[td.SteadyFreeCarrierMonitor(
    ...         center=(0, 0, 0), size=(td.inf, td.inf, 0), name="charge_mnt", unstructured=True
    ...     )],
    ...     center=(0, 0, 0),
    ...     size=(3, 3, 3),
    ...     grid_spec=td.UniformUnstructuredGrid(dl=0.05),
    ...     boundary_spec=[bc_v1, bc_v2],
    ...     analysis_spec=td.IsothermalSteadyChargeDCAnalysis(
    ...         tolerance_settings=td.ChargeToleranceSpec(rel_tol=1e5, abs_tol=3e3, max_iters=400),
    ...         convergence_dv=10),
    ...     )


    Coupling between ``Heat`` and electrical ``Conduction`` simulations is currently limited to 1-way.
    This is specified by defining a heat source of type :class:`HeatFromElectricSource`. With this coupling, joule heating is
    calculated as part  of the solution to a ``Conduction`` simulation and translated into the ``Heat`` simulation.

    Two common scenarios can use this coupling definition:
        1. One in which BCs and sources are specified for both ``Heat`` and ``Conduction`` simulations.
            In this case one mesh will be generated and used for both the ``Conduction`` and ``Heat``
            simulations.
        2. Only heat BCs/sources are provided. In this case, only the ``Heat`` equation will be solved.
            Before the simulation starts, it will try to load the heat source from file so a
            previously run ``Conduction`` simulations must have run previously. Since the Conduction
            and ``Heat`` meshes may differ, an interpolation between them will be performed prior to
            starting the ``Heat`` simulation.

    Additional heat sources can be defined, in which case, they will be added on
    top of the coupling heat source.
    """

    medium: StructureMediumType = pd.Field(
        Medium(),
        title="Background Medium",
        description="Background medium of simulation, defaults to a standard dispersion-less :class:`Medium` if not "
        "specified.",
        discriminator=TYPE_TAG_STR,
    )
    """
    Background medium of simulation, defaults to a standard dispersion-less :class:`Medium` if not specified.
    """

    sources: Tuple[annotate_type(HeatChargeSourceType), ...] = pd.Field(
        (),
        title="Heat and Charge sources",
        description="List of heat and/or charge sources.",
    )

    monitors: Tuple[annotate_type(HeatChargeMonitorType), ...] = pd.Field(
        (),
        title="Monitors",
        description="Monitors in the simulation.",
    )

    boundary_spec: Tuple[annotate_type(Union[HeatChargeBoundarySpec, HeatBoundarySpec]), ...] = (
        pd.Field(
            (),
            title="Boundary Condition Specifications",
            description="List of boundary condition specifications.",
        )
    )
    # NOTE: creating a union with HeatBoundarySpec for backwards compatibility

    grid_spec: UnstructuredGridType = pd.Field(
        title="Grid Specification",
        description="Grid specification for heat-charge simulation.",
        discriminator=TYPE_TAG_STR,
    )

    symmetry: Tuple[ScalarSymmetry, ScalarSymmetry, ScalarSymmetry] = pd.Field(
        (0, 0, 0),
        title="Symmetries",
        description="Tuple of integers defining reflection symmetry across a plane "
        "bisecting the simulation domain normal to the x-, y-, and z-axis "
        "at the simulation center of each axis, respectively. "
        "Each element can be ``0`` (symmetry off) or ``1`` (symmetry on).",
    )

    analysis_spec: AnalysisSpecType = pd.Field(
        None,
        title="Analysis specification.",
        description="The `analysis_spec` is used to validate that the simulation parameters and tolerance settings "
        "are correctly configured as desired by the user.",
    )

    @pd.validator("structures", always=True)
    def check_unsupported_geometries(cls, val):
        """Error if structures contain unsupported yet geometries."""
        for ind, structure in enumerate(val):
            bbox = structure.geometry.bounding_box
            if any(s == 0 for s in bbox.size):
                raise SetupError(
                    f"'HeatSimulation' does not currently support structures with dimensions of zero size ('structures[{ind}]')."
                )
        return val

    @staticmethod
    def _check_cross_solids(objs: Tuple[Box, ...], values: Dict) -> Tuple[int, ...]:
        """Given model dictionary ``values``, check whether objects in list ``objs`` cross
        a ``SolidSpec`` medium.
        """

        # NOTE: when considering Conduction or Charge cases, both conductors and semiconductors
        # will be accepted
        valid_electric_medium = (SemiconductorMedium, ChargeConductorMedium)

        try:
            size = values["size"]
            center = values["center"]
            medium = values["medium"]
            structures = values["structures"]
        except KeyError:
            raise SetupError(
                "Function '_check_cross_solids' assumes dictionary 'values' contains well-defined "
                "'size', 'center',  'medium', and 'structures'. Thus, it should only be used in "
                "validators with @skip_if_fields_missing(['medium', 'center', 'size', 'structures']) "
                "or root validators with option 'skip_on_failure=True'."
            )

        # list of structures including background as a Box()
        structure_bg = Structure(
            geometry=Box(
                size=size,
                center=center,
            ),
            medium=medium,
        )

        total_structures = [structure_bg] + list(structures)

        obj_do_not_cross_solid_idx = []
        obj_do_not_cross_cond_idx = []
        for ind, obj in enumerate(objs):
            if obj.size.count(0.0) == 1:
                # for planar objects we could do a rigorous check
                medium_set = Scene.intersecting_media(obj, total_structures)
                crosses_solid = any(
                    isinstance(medium.heat_spec, SolidMedium) for medium in medium_set
                )
                crosses_elec_spec = any(
                    any([isinstance(medium.charge, medium_i)] for medium_i in valid_electric_medium)
                    for medium in medium_set
                )
            else:
                # approximate check for volumetric objects based on bounding boxes
                # thus, it could still miss a case when there is no data inside the monitor
                crosses_solid = any(
                    obj.intersects(structure.geometry)
                    for structure in total_structures
                    if isinstance(structure.medium.heat_spec, SolidMedium)
                )
                crosses_elec_spec = any(
                    obj.intersects(structure.geometry)
                    for structure in total_structures
                    if any(
                        [isinstance(structure.medium.charge, medium_i)]
                        for medium_i in valid_electric_medium
                    )
                )

            if not crosses_solid:
                obj_do_not_cross_solid_idx.append(ind)
            if not crosses_elec_spec:
                obj_do_not_cross_cond_idx.append(ind)

        return obj_do_not_cross_solid_idx, obj_do_not_cross_cond_idx

    @pd.validator("monitors", always=True)
    @skip_if_fields_missing(["medium", "center", "size", "structures"])
    def _monitors_cross_solids(cls, val, values):
        """Error if monitors does not cross any solid medium."""

        if val is None:
            return val

        failed_solid_idx, failed_elect_idx = cls._check_cross_solids(val, values)

        temp_monitors = [idx for idx, mnt in enumerate(val) if isinstance(mnt, TemperatureMonitor)]
        volt_monitors = [
            idx for idx, mnt in enumerate(val) if isinstance(mnt, SteadyPotentialMonitor)
        ]

        failed_temp_mnt = [idx for idx in temp_monitors if idx in failed_solid_idx]
        failed_volt_mnt = [idx for idx in volt_monitors if idx in failed_elect_idx]

        if len(failed_temp_mnt) > 0:
            monitor_names = [f"'{val[ind].name}'" for ind in failed_temp_mnt]
            raise SetupError(
                f"Monitors {monitor_names} do not cross any solid materials "
                "('heat_spec=SolidSpec(...)'). Temperature distribution is only recorded inside solid "
                "materials. Thus, no information will be recorded in these monitors."
            )

        if len(failed_volt_mnt) > 0:
            monitor_names = [f"'{val[ind].name}'" for ind in failed_volt_mnt]
            raise SetupError(
                f"Monitors {monitor_names} do not cross any conducting materials "
                "('charge=ChargeConductorMedium(...)'). The voltage is only stored inside conducting "
                "materials. Thus, no information will be recorded in these monitors."
            )

        return val

    @pd.root_validator(skip_on_failure=True)
    def check_voltage_array_if_capacitance(cls, values):
        """Make sure an array of voltages has been defined if a
        SteadyCapacitanceMonitor' has been defined"""
        bounday_spec = values["boundary_spec"]
        monitors = values["monitors"]

        is_capacitance_mnt = any(isinstance(mnt, SteadyCapacitanceMonitor) for mnt in monitors)
        voltage_array_present = False
        if is_capacitance_mnt:
            for bc in bounday_spec:
                if isinstance(bc.condition, VoltageBC):
                    if isinstance(bc.condition.source, DCVoltageSource):
                        if isinstance(bc.condition.source.voltage, list) or isinstance(
                            bc.condition.source.voltage, tuple
                        ):
                            if len(bc.condition.source.voltage) > 1:
                                voltage_array_present = True
        if is_capacitance_mnt and not voltage_array_present:
            raise SetupError(
                "Monitors of type 'SteadyCapacitanceMonitor' have been defined but no array of voltages "
                "has been supplied as voltage source, which is required for this type of monitor. "
                "Voltage arrays can be included in a source in this manner: "
                "'VoltageBC(source=DCVoltageSource(voltage=yourArray))'"
            )
        return values

    @pd.validator("size", always=True)
    def check_zero_dim_domain(cls, val, values):
        """Error if heat domain have zero dimensions."""

        dim_names = ["x", "y", "z"]
        zero_dimensions = [False, False, False]
        zero_dim_str = ""
        for n, v in enumerate(val):
            if v == 0:
                zero_dimensions[n] = True
                zero_dim_str += f"{dim_names[n]}- "

        num_zero_dims = np.sum(zero_dimensions)

        if num_zero_dims > 1:
            mssg = f"The current 'HeatChargeSimulation' has zero size along the {zero_dim_str}dimensions. "
            mssg += "Only 2- and 3-D simulations are currently supported."
            raise SetupError(mssg)

        return val

    @pd.validator("boundary_spec", always=True)
    @skip_if_fields_missing(["structures", "medium"])
    def names_exist_bcs(cls, val, values):
        """Error if boundary conditions point to non-existing structures/media."""

        structures = values.get("structures")
        structures_names = {s.name for s in structures}
        mediums_names = {s.medium.name for s in structures}
        mediums_names.add(values.get("medium").name)

        for bc_ind, bc_spec in enumerate(val):
            bc_place = bc_spec.placement
            if isinstance(bc_place, (StructureBoundary, StructureSimulationBoundary)):
                if bc_place.structure not in structures_names:
                    raise SetupError(
                        f"Structure '{bc_place.structure}' provided in "
                        f"'boundary_spec[{bc_ind}].placement' (type '{bc_place.type}') "
                        "is not found among simulation structures."
                    )
            if isinstance(bc_place, (StructureStructureInterface)):
                for struct_name in bc_place.structures:
                    if struct_name and struct_name not in structures_names:
                        raise SetupError(
                            f"Structure '{struct_name}' provided in "
                            f"'boundary_spec[{bc_ind}].placement' (type '{bc_place.type}') "
                            "is not found among simulation structures."
                        )
            if isinstance(bc_place, (MediumMediumInterface)):
                for med_name in bc_place.mediums:
                    if med_name not in mediums_names:
                        raise SetupError(
                            f"Material '{med_name}' provided in "
                            f"'boundary_spec[{bc_ind}].placement' (type '{bc_place.type}') "
                            "is not found among simulation mediums."
                        )
        return val

    @pd.validator("boundary_spec", always=True)
    def check_only_one_voltage_array_provided(cls, val, values):
        """Issue error if more than one voltage array is provided.
        Currently we only allow to sweep over one voltage array.
        """

        array_already_provided = False

        for bc in val:
            if isinstance(bc.condition, VoltageBC):
                voltages = []
                # currently we're only supporting DC BCs, so let's check these values
                if isinstance(bc.condition.source, DCVoltageSource):
                    voltages = bc.condition.source.voltage

                if isinstance(voltages, tuple):
                    if len(voltages) > 1:
                        if not array_already_provided:
                            array_already_provided = True
                        else:
                            raise SetupError(
                                "More than one voltage array has been provided. "
                                "Currently voltage arrays are supported only for one of the BCs."
                            )
        return val

    @pd.root_validator(skip_on_failure=True)
    def check_charge_simulation(cls, values):
        """Makes sure that Charge simulations are set correctly."""

        ChargeMonitorType = (
            SteadyPotentialMonitor,
            SteadyFreeCarrierMonitor,
            SteadyCapacitanceMonitor,
        )

        simulation_types = cls._check_simulation_types(values=values)

        if TCADAnalysisTypes.CHARGE in simulation_types:
            # check that we have at least 2 'VoltageBC's
            boundary_spec = values["boundary_spec"]
            voltage_bcs = 0
            for bc in boundary_spec:
                if isinstance(bc.condition, VoltageBC):
                    voltage_bcs = voltage_bcs + 1
            if voltage_bcs < 2:
                raise SetupError(
                    "Defining a Charge simulation requires the definition of 'VoltageBC' boundaries. "
                    f"So far {voltage_bcs} 'VoltageBC' have been set."
                )

            # check that we have at least one charge monitor
            monitors = values["monitors"]
            if not any(isinstance(mnt, ChargeMonitorType) for mnt in monitors):
                raise SetupError(
                    "Charge simulations require the definition of, at least, one of these monitors: "
                    "'[SteadyPotentialMonitor, SteadyFreeCarrierMonitor, SteadyCapacitanceMonitor]' "
                    "but none have been defined."
                )

            # NOTE: SteadyPotentialMonitor and SteadyFreeCarrierMonitor are only supported
            # for unstructured = True
            for mnt in monitors:
                if isinstance(mnt, SteadyPotentialMonitor) or isinstance(
                    mnt, SteadyFreeCarrierMonitor
                ):
                    if not mnt.unstructured:
                        log.warning(
                            "Currently, Charge simulations support only unstructured monitors. Please set "
                            f"monitor '{mnt.name}' to 'unstructured = True'."
                        )

        return values

    @pd.root_validator(skip_on_failure=True)
    def not_all_neumann(cls, values):
        """Make sure not all BCs are of Neumann type"""

        NeumannBCsHeat = (HeatFluxBC,)
        NeumannBCsCharge = (CurrentBC, InsulatingBC)

        simulation_types = cls._check_simulation_types(values=values)
        bounday_conditions = values["boundary_spec"]

        raise_error = False
        for sim_type in simulation_types:
            if sim_type == TCADAnalysisTypes.HEAT:
                type_bcs = [
                    bc for bc in bounday_conditions if isinstance(bc.condition, HeatBCTypes)
                ]
                if len(type_bcs) == 0 or all(
                    isinstance(bc.condition, NeumannBCsHeat) for bc in type_bcs
                ):
                    raise_error = True
            elif sim_type == TCADAnalysisTypes.CONDUCTION:
                type_bcs = [
                    bc for bc in bounday_conditions if isinstance(bc.condition, ElectricBCTypes)
                ]
                if len(type_bcs) == 0 or all(
                    isinstance(bc.condition, NeumannBCsCharge) for bc in type_bcs
                ):
                    raise_error = True

        names_neumann_Bcs = [BC.__name__ for BC in NeumannBCsHeat]
        names_neumann_Bcs.extend([BC.__name__ for BC in NeumannBCsCharge])
        if raise_error:
            raise SetupError(
                "Current 'HeatChargeSimulation' contains only Neumann-type boundary conditions. "
                "Steady-state solution is undefined in this case. "
                f"Current Neumann BCs are {names_neumann_Bcs}"
            )

        return values

    @pd.validator("grid_spec", always=True)
    @skip_if_fields_missing(["structures"])
    def names_exist_grid_spec(cls, val, values):
        """Warn if 'UniformUnstructuredGrid' points at a non-existing structure."""

        structures = values.get("structures")
        structures_names = {s.name for s in structures}

        for structure_name in val.non_refined_structures:
            if structure_name not in structures_names:
                log.warning(
                    f"Structure '{structure_name}' listed as a non-refined structure in "
                    "'HeatChargeSimulation.grid_spec' is not present in 'HeatChargeSimulation.structures'"
                )

        return val

    @pd.validator("grid_spec", always=True)
    def warn_if_minimal_mesh_size_override(cls, val, values):
        """Warn if minimal mesh size limit overrides desired mesh size."""

        max_size = np.max(values.get("size"))
        min_dl = val.relative_min_dl * max_size

        if isinstance(val, UniformUnstructuredGrid):
            desired_min_dl = val.dl
        if isinstance(val, DistanceUnstructuredGrid):
            desired_min_dl = min(val.dl_interface, val.dl_bulk)

        if desired_min_dl < min_dl:
            log.warning(
                f"The resulting limit for minimal mesh size from parameter 'relative_min_dl={val.relative_min_dl}' is {min_dl}, while provided mesh size in 'grid_spec' is {desired_min_dl}. "
                "Consider lowering parameter 'relative_min_dl' if a finer grid is required."
            )

        return val

    @pd.validator("sources", always=True)
    @skip_if_fields_missing(["structures"])
    def names_exist_sources(cls, val, values):
        """Error if a heat-charge source point to non-existing structures."""
        structures = values.get("structures")
        structures_names = {s.name for s in structures}

        sources = [s for s in val if not isinstance(s, HeatFromElectricSource)]

        for source in sources:
            for name in source.structures:
                if name not in structures_names:
                    raise SetupError(
                        f"Structure '{name}' provided in a '{source.type}' "
                        "is not found among simulation structures."
                    )
        return val

    @pd.root_validator(skip_on_failure=True)
    def check_medium_specs(cls, values):
        """Error if no appropriate specs."""

        sim_box = (
            Box(
                size=values.get("size"),
                center=values.get("center"),
            ),
        )

        failed_solid_idx, failed_elect_idx = cls._check_cross_solids(sim_box, values)

        simulation_types = cls._check_simulation_types(values=values)

        for sim_type in simulation_types:
            if sim_type == TCADAnalysisTypes.HEAT:
                if len(failed_solid_idx) > 0:
                    raise SetupError(
                        "No solid materials ('SolidSpec') are detected in heat simulation. Solution domain is empty."
                    )
            elif sim_type == TCADAnalysisTypes.CONDUCTION:
                if len(failed_elect_idx) > 0:
                    raise SetupError(
                        "No conducting materials ('ChargeConductorMedium') are detected in conduction simulation. Solution domain is empty."
                    )

        return values

    @staticmethod
    def _check_if_semiconductor_present(structures) -> bool:
        """Checks whether the simulation object can run a Charge simulation."""

        charge_sim = False

        # make sure mediums with doping have been defined
        for structure in structures:
            if isinstance(structure.medium, MultiPhysicsMedium):
                if structure.medium.charge is not None:
                    if isinstance(structure.medium.charge, SemiconductorMedium):
                        return True
        return charge_sim

    @staticmethod
    def _check_simulation_types(
        values: Dict,
        HeatBCTypes=HeatBCTypes,
        ElectricBCTypes=ElectricBCTypes,
        HeatSourceTypes=HeatSourceTypes,
    ) -> list[TCADAnalysisTypes]:
        """Given model dictionary ``values``, check the type of simulations to be run
        based on BCs and sources.
        """
        simulation_types = []

        boundaries = list(values["boundary_spec"])
        sources = list(values["sources"])

        structures = list(values["structures"])
        semiconductor_present = HeatChargeSimulation._check_if_semiconductor_present(
            structures=structures
        )
        if semiconductor_present:
            simulation_types.append(TCADAnalysisTypes.CHARGE)

        for boundary in boundaries:
            if isinstance(boundary.condition, HeatBCTypes):
                simulation_types.append(TCADAnalysisTypes.HEAT)
            if isinstance(boundary.condition, ElectricBCTypes):
                # for the time being, assume tha the simulation will be of
                # type CHARGE if we have semiconductors
                if semiconductor_present:
                    simulation_types.append(TCADAnalysisTypes.CHARGE)
                else:
                    simulation_types.append(TCADAnalysisTypes.CONDUCTION)

        for source in sources:
            if isinstance(source, HeatSourceTypes):
                simulation_types.append(TCADAnalysisTypes.HEAT)

        return set(simulation_types)

    @pd.root_validator(skip_on_failure=True)
    def check_coupling_source_can_be_applied(cls, values):
        """Error if material doesn't have the right specifications"""

        HeatSourceTypes_noCoupling = (UniformHeatSource, HeatSource)

        simulation_types = cls._check_simulation_types(
            values, HeatSourceTypes=HeatSourceTypes_noCoupling
        )
        simulation_types = list(simulation_types)

        sources = list(values["sources"])

        for source in sources:
            if isinstance(source, HeatFromElectricSource) and len(simulation_types) < 2:
                raise SetupError(
                    f"Using 'HeatFromElectricSource' requires the definition of both "
                    f"{TCADAnalysisTypes.CONDUCTION.name} and {TCADAnalysisTypes.HEAT.name}. "
                    f"The current simulation setup contains only conditions of type {simulation_types[0].name}"
                )

        return values

    @pd.root_validator(skip_on_failure=True)
    def estimate_charge_mesh_size(cls, values):
        """Make an estimate of the mesh size and raise a warning if too big.
        NOTE: this is a very rough estimate. The back-end will actually stop
        execution based on actual node-count."""

        if TCADAnalysisTypes.CHARGE not in cls._check_simulation_types(values=values):
            return values

        # let's raise a warning if the estimate is larger than 2M nodes
        max_nodes = 2e6
        nodes_estimate = 0

        structures = values["structures"]
        grid_spec = values["grid_spec"]

        non_refined_structures = grid_spec.non_refined_structures

        if isinstance(grid_spec, UniformUnstructuredGrid):
            dl_min = grid_spec.dl
            dl_max = dl_min
        elif isinstance(grid_spec, DistanceUnstructuredGrid):
            dl_min = grid_spec.dl_interface
            dl_max = grid_spec.dl_bulk

        for struct in structures:
            name = struct.name
            bounds = struct.geometry.bounds
            dl = dl_min
            if name in non_refined_structures:
                dl = dl_max
            nodes_structure = 1
            for coord_min, coord_max in zip(bounds[0], bounds[1]):
                if (
                    (coord_min != coord_max)
                    and (np.abs(coord_min) != np.inf)
                    and (np.abs(coord_max) != np.inf)
                ):
                    nodes_structure = nodes_structure * (coord_max - coord_min) / dl

            nodes_estimate = nodes_estimate + nodes_structure

        if nodes_estimate > max_nodes:
            log.warning(
                "WARNING: It has been estimated the mesh to be bigger than the currently "
                "supported mesh size for Charge simulations. The simulation may be "
                "submitted but if the maximum number of nodes is indeed exceeded "
                "the pipeline will be stopped. If this happens the grid specification "
                "may need to be modified."
            )
        return values

    @equal_aspect
    @add_ax_if_none
    def plot_property(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        alpha: float = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        property: str = "heat_conductivity",
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
    ) -> Ax:
        """Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        property : str = "heat_conductivity"
            Specified the type of simulation for which the plot will be tailored.
            Options are ["heat_conductivity", "electric_conductivity", "source"]
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        hlim, vlim = Scene._get_plot_lims(
            bounds=self.simulation_bounds, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )

        cbar_cond = True

        simulation_types = self._get_simulation_types()
        if property == "source" and len(simulation_types) > 1:
            raise ValueError(
                "'plot_property' must be called with argument 'property' in "
                "'HeatChargeSimulations' with multiple physics, i.e., a 'HeatChargeSimulation' "
                f"with both {TCADAnalysisTypes.HEAT.name} and "
                f"{TCADAnalysisTypes.CONDUCTION.name} simulation properties."
            )
        if len(simulation_types) == 1:
            if (
                property == "heat_conductivity" and TCADAnalysisTypes.CONDUCTION in simulation_types
            ) or (
                property == "electric_conductivity" and TCADAnalysisTypes.HEAT in simulation_types
            ):
                raise ValueError(
                    f"'property' in 'plot_property()' was defined as {property} but the "
                    f"simulation is of type {simulation_types[0]}."
                )

        if property != "source":
            ax = self.scene.plot_heat_charge_property(
                ax=ax,
                x=x,
                y=y,
                z=z,
                cbar=cbar_cond,
                alpha=alpha,
                hlim=hlim,
                vlim=vlim,
                property=property,
            )
        ax = self.plot_sources(
            ax=ax, x=x, y=y, z=z, property=property, alpha=source_alpha, hlim=hlim, vlim=vlim
        )
        ax = self.plot_monitors(ax=ax, x=x, y=y, z=z, alpha=monitor_alpha, hlim=hlim, vlim=vlim)
        ax = self.plot_boundaries(ax=ax, x=x, y=y, z=z, property=property)
        ax = Scene._set_plot_bounds(
            bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim
        )
        ax = self.plot_symmetries(ax=ax, x=x, y=y, z=z, hlim=hlim, vlim=vlim)

        if property == "source":
            self._add_source_cbar(ax=ax, property=property)
        return ax

    @equal_aspect
    @add_ax_if_none
    def plot_heat_conductivity(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        ax: Ax = None,
        alpha: float = None,
        source_alpha: float = None,
        monitor_alpha: float = None,
        colorbar: str = "conductivity",
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
        **kwargs,
    ) -> Ax:
        """
        DEPRECATED: Method added for backwards compatibility with :class:`HeatSimulation.plot_heat_conductivity`.
        Plot each of simulation's components on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        alpha : float = None
            Opacity of the structures being plotted.
            Defaults to the structure default alpha.
        source_alpha : float = None
            Opacity of the sources. If ``None``, uses Tidy3d default.
        monitor_alpha : float = None
            Opacity of the monitors. If ``None``, uses Tidy3d default.
        colorbar: str = "conductivity"
            Display colorbar for thermal conductivity ("conductivity") or heat source rate
            ("source").
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        log.warning(
            """This function `plot_heat_conductivity` is 
            deprecated and will be discontinued. In its place you can use
            `plot_property(property="heat_conductivity")`"""
        )

        plot_type = "heat_conductivity"
        if colorbar == "conductivity":
            plot_type = "heat_conductivity"
        elif colorbar == "source":
            plot_type = "source"

        return self.plot_property(
            x=x,
            y=y,
            z=z,
            ax=ax,
            alpha=alpha,
            source_alpha=source_alpha,
            monitor_alpha=monitor_alpha,
            property=plot_type,
            hlim=hlim,
            vlim=vlim,
        )

    @equal_aspect
    @add_ax_if_none
    def plot_boundaries(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        property: str = "heat_conductivity",
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's boundary conditions on a plane defined by one nonzero x,y,z
        coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        property : str = None
            Specified the type of simulation for which the plot will be tailored.
            Options are ["heat_conductivity", "electric_conductivity"]
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        # get structure list
        structures = [self.simulation_structure]
        structures += list(self.structures)

        # construct slicing plane
        axis, position = Box.parse_xyz_kwargs(x=x, y=y, z=z)
        center = Box.unpop_axis(position, (0, 0), axis=axis)
        size = Box.unpop_axis(0, (inf, inf), axis=axis)
        plane = Box(center=center, size=size)

        # get boundary conditions in the plane
        boundaries = self._construct_heat_charge_boundaries(
            structures=structures,
            plane=plane,
            boundary_spec=self.boundary_spec,
        )

        # plot boundary conditions
        if property == "heat_conductivity" or property == "source":
            new_boundaries = [(b, s) for b, s in boundaries if isinstance(b.condition, HeatBCTypes)]
        elif property == "electric_conductivity":
            new_boundaries = [
                (b, s) for b, s in boundaries if isinstance(b.condition, ElectricBCTypes)
            ]

        for bc_spec, shape in new_boundaries:
            ax = self._plot_boundary_condition(shape=shape, boundary_spec=bc_spec, ax=ax)

        # clean up the axis display
        ax = self.add_ax_lims(axis=axis, ax=ax)
        ax = Scene._set_plot_bounds(bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z)
        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )

        return ax

    def _get_bc_plot_params(self, boundary_spec: HeatChargeBoundarySpec) -> PlotParams:
        """Constructs the plot parameters for given boundary conditions."""

        plot_params = plot_params_heat_bc
        condition = boundary_spec.condition

        if isinstance(condition, TemperatureBC):
            plot_params = plot_params.updated_copy(facecolor=HEAT_BC_COLOR_TEMPERATURE)
        elif isinstance(condition, HeatFluxBC):
            plot_params = plot_params.updated_copy(facecolor=HEAT_BC_COLOR_FLUX)
        elif isinstance(condition, ConvectionBC):
            plot_params = plot_params.updated_copy(facecolor=HEAT_BC_COLOR_CONVECTION)
        elif isinstance(condition, InsulatingBC):
            plot_params = plot_params.updated_copy(facecolor=CHARGE_BC_INSULATOR)

        return plot_params

    def _plot_boundary_condition(
        self, shape: Shapely, boundary_spec: HeatChargeBoundarySpec, ax: Ax
    ) -> Ax:
        """Plot a structure's cross section shape for a given boundary condition."""
        plot_params_bc = self._get_bc_plot_params(boundary_spec=boundary_spec)
        ax = self.plot_shape(shape=shape, plot_params=plot_params_bc, ax=ax)
        return ax

    @staticmethod
    def _structure_to_bc_spec_map(
        plane: Box,
        structures: Tuple[Structure, ...],
        boundary_spec: Tuple[HeatChargeBoundarySpec, ...],
    ) -> Dict[str, HeatChargeBoundarySpec]:
        """Construct structure name to bc spec inverse mapping. One structure may correspond to
        multiple boundary conditions."""

        named_structures_present = {structure.name for structure in structures if structure.name}

        struct_to_bc_spec = {}
        for bc_spec in boundary_spec:
            bc_place = bc_spec.placement
            if (
                isinstance(bc_place, (StructureBoundary, StructureSimulationBoundary))
                and bc_place.structure in named_structures_present
            ):
                if bc_place.structure in struct_to_bc_spec:
                    struct_to_bc_spec[bc_place.structure] += [bc_spec]
                else:
                    struct_to_bc_spec[bc_place.structure] = [bc_spec]

            if isinstance(bc_place, StructureStructureInterface):
                for structure in bc_place.structures:
                    if structure in named_structures_present:
                        if structure in struct_to_bc_spec:
                            struct_to_bc_spec[structure] += [bc_spec]
                        else:
                            struct_to_bc_spec[structure] = [bc_spec]

            if isinstance(bc_place, SimulationBoundary):
                struct_to_bc_spec[HEAT_CHARGE_BACK_STRUCTURE_STR] = [bc_spec]

        return struct_to_bc_spec

    @staticmethod
    def _medium_to_bc_spec_map(
        plane: Box,
        structures: Tuple[Structure, ...],
        boundary_spec: Tuple[HeatChargeBoundarySpec, ...],
    ) -> Dict[str, HeatChargeBoundarySpec]:
        """Construct medium name to bc spec inverse mapping. One medium may correspond to
        multiple boundary conditions."""

        named_mediums_present = {
            structure.medium.name for structure in structures if structure.medium.name
        }

        med_to_bc_spec = {}
        for bc_spec in boundary_spec:
            bc_place = bc_spec.placement
            if isinstance(bc_place, MediumMediumInterface):
                for med in bc_place.mediums:
                    if med in named_mediums_present:
                        if med in med_to_bc_spec:
                            med_to_bc_spec[med] += [bc_spec]
                        else:
                            med_to_bc_spec[med] = [bc_spec]

        return med_to_bc_spec

    @staticmethod
    def _construct_forward_boundaries(
        shapes: Tuple[Tuple[str, str, Shapely, Tuple[float, float, float, float]], ...],
        struct_to_bc_spec: Dict[str, HeatChargeBoundarySpec],
        med_to_bc_spec: Dict[str, HeatChargeBoundarySpec],
        background_structure_shape: Shapely,
    ) -> Tuple[Tuple[HeatChargeBoundarySpec, Shapely], ...]:
        """Construct Simulation, StructureSimulation, Structure, and MediumMedium boundaries."""

        # forward foop to take care of Simulation, StructureSimulation, Structure,
        # and MediumMediums
        boundaries = []  # bc_spec, structure name, shape, bounds
        background_shapes = []
        for name, medium, shape, bounds in shapes:
            # intersect existing boundaries (both structure based and medium based)
            for index, (_bc_spec, _name, _bdry, _bounds) in enumerate(boundaries):
                # simulation bc is overridden only by StructureSimulationBoundary
                if isinstance(_bc_spec.placement, SimulationBoundary):
                    if name not in struct_to_bc_spec:
                        continue
                    if any(
                        not isinstance(bc_spec.placement, StructureSimulationBoundary)
                        for bc_spec in struct_to_bc_spec[name]
                    ):
                        continue

                if Box._do_not_intersect(bounds, _bounds, shape, _bdry):
                    continue

                diff_shape = _bdry - shape

                boundaries[index] = (_bc_spec, _name, diff_shape, diff_shape.bounds)

            # create new structure based boundary

            if name in struct_to_bc_spec:
                for bc_spec in struct_to_bc_spec[name]:
                    if isinstance(bc_spec.placement, StructureBoundary):
                        bdry = shape.exterior
                        bdry = bdry.intersection(background_structure_shape)
                        boundaries.append((bc_spec, name, bdry, bdry.bounds))

                    if isinstance(bc_spec.placement, SimulationBoundary):
                        boundaries.append((bc_spec, name, shape.exterior, shape.exterior.bounds))

                    if isinstance(bc_spec.placement, StructureSimulationBoundary):
                        bdry = background_structure_shape.exterior
                        bdry = bdry.intersection(shape)
                        boundaries.append((bc_spec, name, bdry, bdry.bounds))

            # create new medium based boundary, and cut or merge relevant background shapes

            # loop through background_shapes (note: all background are non-intersecting or merged)
            # this is similar to _filter_structures_plane but only mediums participating in BCs
            # are tracked
            for index, (_medium, _shape, _bounds) in enumerate(background_shapes):
                if Box._do_not_intersect(bounds, _bounds, shape, _shape):
                    continue

                diff_shape = _shape - shape

                # different medium, remove intersection from background shape
                if medium != _medium and len(diff_shape.bounds) > 0:
                    background_shapes[index] = (_medium, diff_shape, diff_shape.bounds)

                    # in case when there is a bc between two media
                    # create a new boundary segment
                    for bc_spec in med_to_bc_spec[_medium.name]:
                        if medium.name in bc_spec.placement.mediums:
                            bdry = shape.exterior.intersection(_shape)
                            bdry = bdry.intersection(background_structure_shape)
                            boundaries.append((bc_spec, name, bdry, bdry.bounds))

                # same medium, add diff shape to this shape and mark background shape for removal
                # note: this only happens if this medium is listed in BCs
                else:
                    shape = shape | diff_shape
                    background_shapes[index] = None

            # after doing this with all background shapes, add this shape to the background
            # but only if this medium is listed in BCs
            if medium.name in med_to_bc_spec:
                background_shapes.append((medium, shape, shape.bounds))

            # remove any existing background shapes that have been marked as 'None'
            background_shapes = [b for b in background_shapes if b is not None]

        # filter out empty geometries
        boundaries = [(bc_spec, bdry) for (bc_spec, name, bdry, _) in boundaries if bdry]

        return boundaries

    @staticmethod
    def _construct_reverse_boundaries(
        shapes: Tuple[Tuple[str, str, Shapely, Bound], ...],
        struct_to_bc_spec: Dict[str, HeatChargeBoundarySpec],
        background_structure_shape: Shapely,
    ) -> Tuple[Tuple[HeatChargeBoundarySpec, Shapely], ...]:
        """Construct StructureStructure boundaries."""

        # backward foop to take care of StructureStructure
        # we do it in this way because we define the boundary between
        # two overlapping structures A and B, where A comes before B, as
        # boundary(B) intersected by A
        # So, in this loop as we go backwards through the structures we:
        # - (1) when come upon B, create boundary(B)
        # - (2) cut away from it by other structures
        # - (3) when come upon A, intersect it with A and mark it as complete,
        #   that is, no more further modifications
        boundaries_reverse = []

        for name, _, shape, bounds in shapes[:0:-1]:
            minx, miny, maxx, maxy = bounds

            # intersect existing boundaries
            for index, (_bc_spec, _name, _bdry, _bounds, _completed) in enumerate(
                boundaries_reverse
            ):
                if not _completed:
                    if Box._do_not_intersect(bounds, _bounds, shape, _bdry):
                        continue

                    # event (3) from above
                    if name in _bc_spec.placement.structures:
                        new_bdry = _bdry.intersection(shape)
                        boundaries_reverse[index] = (
                            _bc_spec,
                            _name,
                            new_bdry,
                            new_bdry.bounds,
                            True,
                        )

                    # event (2) from above
                    else:
                        new_bdry = _bdry - shape
                        boundaries_reverse[index] = (
                            _bc_spec,
                            _name,
                            new_bdry,
                            new_bdry.bounds,
                            _completed,
                        )

            # create new boundary (event (1) from above)
            if name in struct_to_bc_spec:
                for bc_spec in struct_to_bc_spec[name]:
                    if isinstance(bc_spec.placement, StructureStructureInterface):
                        bdry = shape.exterior
                        bdry = bdry.intersection(background_structure_shape)
                        boundaries_reverse.append((bc_spec, name, bdry, bdry.bounds, False))

        # filter and append completed boundaries to main list
        filtered_boundaries = []
        for bc_spec, _, bdry, _, is_completed in boundaries_reverse:
            if bdry and is_completed:
                filtered_boundaries.append((bc_spec, bdry))

        return filtered_boundaries

    @staticmethod
    def _construct_heat_charge_boundaries(
        structures: List[Structure],
        plane: Box,
        boundary_spec: List[HeatChargeBoundarySpec],
    ) -> List[Tuple[HeatChargeBoundarySpec, Shapely]]:
        """Compute list of boundary lines to plot on plane.

        Parameters
        ----------
        structures : List[:class:`.Structure`]
            list of structures to filter on the plane.
        plane : :class:`.Box`
            target plane.
        boundary_spec : List[HeatBoundarySpec]
            list of boundary conditions associated with structures.

        Returns
        -------
        List[Tuple[:class:`.HeatBoundarySpec`, shapely.geometry.base.BaseGeometry]]
            List of boundary lines and boundary conditions on the plane after merging.
        """

        # get structures in the plane and present named structures and media
        shapes = []  # structure name, structure medium, shape, bounds
        for structure in structures:
            # get list of Shapely shapes that intersect at the plane
            shapes_plane = plane.intersections_with(structure.geometry)

            # append each of them and their medium information to the list of shapes
            for shape in shapes_plane:
                shapes.append((structure.name, structure.medium, shape, shape.bounds))

        background_structure_shape = shapes[0][2]

        # construct an inverse mapping structure -> bc for present structures
        struct_to_bc_spec = HeatChargeSimulation._structure_to_bc_spec_map(
            plane=plane, structures=structures, boundary_spec=boundary_spec
        )

        # construct an inverse mapping medium -> bc for present mediums
        med_to_bc_spec = HeatChargeSimulation._medium_to_bc_spec_map(
            plane=plane, structures=structures, boundary_spec=boundary_spec
        )

        # construct boundaries in 2 passes:

        # 1. forward foop to take care of Simulation, StructureSimulation, Structure,
        # and MediumMediums
        boundaries = HeatChargeSimulation._construct_forward_boundaries(
            shapes=shapes,
            struct_to_bc_spec=struct_to_bc_spec,
            med_to_bc_spec=med_to_bc_spec,
            background_structure_shape=background_structure_shape,
        )

        # 2. reverse loop: construct structure-structure boundary
        struct_struct_boundaries = HeatChargeSimulation._construct_reverse_boundaries(
            shapes=shapes,
            struct_to_bc_spec=struct_to_bc_spec,
            background_structure_shape=background_structure_shape,
        )

        return boundaries + struct_struct_boundaries

    @equal_aspect
    @add_ax_if_none
    def plot_sources(
        self,
        x: float = None,
        y: float = None,
        z: float = None,
        property: str = "heat_conductivity",
        hlim: Tuple[float, float] = None,
        vlim: Tuple[float, float] = None,
        alpha: float = None,
        ax: Ax = None,
    ) -> Ax:
        """Plot each of simulation's sources on a plane defined by one nonzero x,y,z coordinate.

        Parameters
        ----------
        x : float = None
            position of plane in x direction, only one of x, y, z must be specified to define plane.
        y : float = None
            position of plane in y direction, only one of x, y, z must be specified to define plane.
        z : float = None
            position of plane in z direction, only one of x, y, z must be specified to define plane.
        property : str = None
            Specified the type of simulation for which the plot will be tailored.
            Options are ["heat_conductivity", "electric_conductivity"]
        hlim : Tuple[float, float] = None
            The x range if plotting on xy or xz planes, y range if plotting on yz plane.
        vlim : Tuple[float, float] = None
            The z range if plotting on xz or yz planes, y plane if plotting on xy plane.
        alpha : float = None
            Opacity of the sources, If ``None`` uses Tidy3d default.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """

        # background can't have source, so no need to add background structure
        structures = self.structures

        # alpha is None just means plot without any transparency
        if alpha is None:
            alpha = 1

        if alpha <= 0:
            return ax

        # get appropriate sources
        if property == "heat_conductivity" or property == "source":
            source_list = [s for s in self.sources if isinstance(s, HeatSourceTypes)]
        elif property == "electric_conductivity":
            source_list = [s for s in self.sources if isinstance(s, ChargeSourceTypes)]

        # distribute source where there are assigned
        structure_source_map = {}
        for source in source_list:
            if not isinstance(source, GlobalHeatChargeSource):
                for name in source.structures:
                    structure_source_map[name] = source

        source_list = [structure_source_map.get(structure.name, None) for structure in structures]

        axis, position = Box.parse_xyz_kwargs(x=x, y=y, z=z)
        center = Box.unpop_axis(position, (0, 0), axis=axis)
        size = Box.unpop_axis(0, (inf, inf), axis=axis)
        plane = Box(center=center, size=size)

        source_shapes = self.scene._filter_structures_plane(
            structures=structures, plane=plane, property_list=source_list
        )

        source_min, source_max = self.source_bounds(property=property)
        for source, shape in source_shapes:
            if source is not None:
                ax = self._plot_shape_structure_source(
                    alpha=alpha,
                    source=source,
                    source_min=source_min,
                    source_max=source_max,
                    shape=shape,
                    ax=ax,
                )

        # clean up the axis display
        ax = self.add_ax_lims(axis=axis, ax=ax)
        ax = Scene._set_plot_bounds(bounds=self.simulation_bounds, ax=ax, x=x, y=y, z=z)
        # Add the default axis labels, tick labels, and title
        ax = Box.add_ax_labels_and_title(
            ax=ax, x=x, y=y, z=z, plot_length_units=self.plot_length_units
        )
        return ax

    def _add_source_cbar(self, ax: Ax, property: str = "heat_conductivity"):
        """Add colorbar for heat sources."""
        source_min, source_max = self.source_bounds(property=property)
        self.scene._add_cbar(
            vmin=source_min,
            vmax=source_max,
            label=f"Volumetric heat rate ({VOLUMETRIC_HEAT_RATE})",
            cmap=HEAT_SOURCE_CMAP,
            ax=ax,
        )

    def _safe_float_conversion(self, string) -> float:
        """Function to deal with failed string2float conversion when using
        expressions in 'HeatSource'"""
        try:
            return float(string)
        except ValueError:
            return None

    def source_bounds(self, property: str = "heat_conductivity") -> Tuple[float, float]:
        """Compute range of heat sources present in the simulation."""

        if property == "heat_conductivity" or property == "source":
            rate_list = [
                self._safe_float_conversion(source.rate)
                for source in self.sources
                if isinstance(source, HeatSource)
            ]
        elif property == "electric_conductivity":
            rate_list = [
                self._safe_float_conversion(source.rate)
                for source in self.sources
                if isinstance(source, ChargeSourceTypes)
            ]  # this is currently an empty list

        rate_list.append(0)
        rate_min = min(rate_list)
        rate_max = max(rate_list)
        return rate_min, rate_max

    def _get_structure_source_plot_params(
        self,
        source: HeatChargeSourceType,
        source_min: float,
        source_max: float,
        alpha: float = None,
    ) -> PlotParams:
        """Constructs the plot parameters for a given medium in simulation.plot_eps()."""

        plot_params = plot_params_heat_source
        if alpha is not None:
            plot_params = plot_params.copy(update={"alpha": alpha})

        if isinstance(source, HeatSource):
            rate = self._safe_float_conversion(source.rate)
            if rate is not None:
                delta_rate = rate - source_min
                delta_rate_max = source_max - source_min + 1e-5
                rate_fraction = delta_rate / delta_rate_max
                cmap = colormaps[HEAT_SOURCE_CMAP]
                rgba = cmap(rate_fraction)
                plot_params = plot_params.copy(update={"edgecolor": rgba})

        return plot_params

    def _plot_shape_structure_source(
        self,
        source: HeatChargeSourceType,
        shape: Shapely,
        source_min: float,
        source_max: float,
        ax: Ax,
        alpha: float = None,
    ) -> Ax:
        """Plot a structure's cross section shape for a given medium, grayscale for permittivity."""
        plot_params = self._get_structure_source_plot_params(
            source=source,
            source_min=source_min,
            source_max=source_max,
            alpha=alpha,
        )
        ax = self.plot_shape(shape=shape, plot_params=plot_params, ax=ax)
        return ax

    @classmethod
    def from_scene(cls, scene: Scene, **kwargs) -> HeatChargeSimulation:
        """Create a simulation from a :class:`.Scene` instance. Must provide additional parameters
        to define a valid simulation (for example, ``size``, ``grid_spec``, etc).

        Parameters
        ----------
        scene : :class:`.Scene`
            Scene containing structures information.
        **kwargs
            Other arguments

        Example
        -------
        >>> from tidy3d import Scene, Medium, Box, Structure, UniformUnstructuredGrid
        >>> box = Structure(
        ...     geometry=Box(center=(0, 0, 0), size=(1, 2, 3)),
        ...     medium=Medium(permittivity=5),
        ...     name="box"
        ... )
        >>> scene = Scene(
        ...     structures=[box],
        ...     medium=Medium(
        ...         permittivity=3,
        ...         heat_spec=SolidMedium(
        ...             conductivity=1, capacity=1,
        ...         ),
        ...     ),
        ... )
        >>> sim = HeatChargeSimulation.from_scene(
        ...     scene=scene,
        ...     center=(0, 0, 0),
        ...     size=(5, 6, 7),
        ...     grid_spec=UniformUnstructuredGrid(dl=0.4),
        ...     boundary_spec=[
        ...         HeatChargeBoundarySpec(
        ...             placement=StructureBoundary(structure="box"),
        ...             condition=TemperatureBC(temperature=500),
        ...         )
        ...     ],
        ... )
        """

        return cls(
            structures=scene.structures,
            medium=scene.medium,
            **kwargs,
        )

    def _get_simulation_types(self) -> list[TCADAnalysisTypes]:
        """
        Checks through BCs and sources and returns the
        types of simulations.
        """
        simulation_types = []

        # NOTE: for the time being, if a simulation has SemiconductorMedium
        # then we consider it of being a 'TCADAnalysisTypes.CHARGE'
        if self._check_if_semiconductor_present(self.structures):
            return [TCADAnalysisTypes.CHARGE]

        heat_source_present = any(isinstance(s, HeatSourceTypes) for s in self.sources)

        heat_BCs_present = any(isinstance(bc.condition, HeatBCTypes) for bc in self.boundary_spec)

        if heat_source_present and not heat_BCs_present:
            raise SetupError("Heat sources defined but no heat BCs present.")
        elif heat_BCs_present or heat_source_present:
            simulation_types.append(TCADAnalysisTypes.HEAT)

        # check for conduction simulation
        electric_spec_present = any(
            structure.medium.charge is not None for structure in self.structures
        )

        electric_BCs_present = any(
            isinstance(bc.condition, ElectricBCTypes) for bc in self.boundary_spec
        )

        if electric_BCs_present and not electric_spec_present:
            raise SetupError(
                "Electric BC were specified but no structure in the simulation has "
                "a defined '.medium.charge'. Structures with "
                "'.medium.charge=None' are treated as insulators, thus, "
                "the solution domain is empty."
            )
        elif electric_BCs_present and electric_spec_present:
            simulation_types.append(TCADAnalysisTypes.CONDUCTION)

        return simulation_types

    def _useHeatSourceFromConductionSim(self):
        """Returns True if 'HeatFromElectricSource' has been defined."""

        return any(isinstance(source, HeatFromElectricSource) for source in self.sources)
